from v2simux_gui.common import *
from v2simux import AMAP_KEY_FILE
from .utils import *
from .controls import ScrollableTreeView


_L = LangLib.Load(__file__)


class CSCSVEditor(Frame):
    def __init__(self, master, down_worker, file:str="", **kwargs):
        super().__init__(master, **kwargs)

        self._Q = EventQueue(self)
        self._Q.register("loaded", lambda: None)

        if file:
            self.file = file
        else:
            self.file = ""
        self.down_wk = down_worker
        self.tree = ScrollableTreeView(self) 
        self.tree['show'] = 'headings'
        self.tree["columns"] = ("ID", "Address", "X", "Y")
        self.tree.column("ID", width=120, stretch=NO)
        self.tree.column("X", width=100, stretch=NO)
        self.tree.column("Y", width=100, stretch=NO)
        self.tree.column("Address", width=180, stretch=YES)
        
        self.tree.heading("ID", text=_L["CSCSV_ID"])
        self.tree.heading("X", text=_L["CSCSV_X"])
        self.tree.heading("Y", text=_L["CSCSV_Y"])
        self.tree.heading("Address", text=_L["CSCSV_ADDR"])
        self.tree.pack(fill="both", expand=True)

        self.lb_cnt = Label(self, text=_L["LB_COUNT"].format(0))
        self.lb_cnt.pack(fill="x", expand=False)

        self.panel = Frame(self)
        self.panel.pack(fill="x", expand=False)
        self.btn_down = Button(self.panel, text=_L["CSCSV_DOWNLOAD"], command=self.down)
        self.btn_down.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.lb_amapkey = Label(self.panel, text=_L["CSCSV_KEY"])
        self.lb_amapkey.grid(row=0, column=1, padx=3, pady=3, sticky="w")
        self.entry_amapkey = Entry(self.panel, width=50)
        self.entry_amapkey.grid(row=0, column=2, columnspan=2, padx=3, pady=3, sticky="w")

        if Path(AMAP_KEY_FILE).exists():
            with open(AMAP_KEY_FILE, "r") as f:
                self.entry_amapkey.insert(0, f.read().strip())
        
    def down(self):
        if MB.askyesno(_L["CSCSV_CONFIRM_TITLE"], _L["CSCSV_CONFIRM"]):
            with open(AMAP_KEY_FILE, "w") as f:
                f.write(self.entry_amapkey.get().strip())
            self.down_wk()
    
    def __readfile(self, file:str, encoding:str):
        try:
            with open(file, "r", encoding=encoding) as f:
                f.readline()
                lines = f.readlines()
            return lines
        except UnicodeDecodeError:
            return None

    def __load(self, file:str):
        encodings = ['utf-8', 'gbk']
        lines = None
        for enc in encodings:
            try:
                lines = self.__readfile(file, enc)
            except Exception as e:
                showerr(_L["ERROR_LOADING"].format(file, e))
                return
            if lines is not None: break
        else:
            showerr(_L["ERROR_LOADING"].format(file, _L["UKN_ENCODING"]))
            return
            
        self.file = file
        self.lb_cnt.config(text=_L["LB_COUNT"].format(len(lines) - 1))
        self.tree.clear()
        for i, cs in enumerate(lines, start=2):
            vals = cs.strip().split(',')
            if len(vals) != 4:
                print(_L["INVALID_LINE_IN_CSCSV"].format(i, cs))
            self._Q.delegate(self.tree.insert, "", "end", values=tuple(vals))

    def load(self, file:str):
        self._Q.submit("loaded", self.__load, file)
    
    def clear(self):
        self.lb_cnt.config(text=_L["LB_COUNT"].format(0))
        self.tree.clear()


from pathlib import Path
from typing import Dict, Optional
from v2simux import (
    Lang, PluginPool, StaPool,
    load_external_components,
    simulate_single,
    get_sim_params,
)
import threading, sys, logging, platform
from feasytools import ArgChecker, KeyNotSpecifiedError


def error_exit(err=None, print_help: bool = False):
    if err:
        if isinstance(err, KeyNotSpecifiedError):
            print(Lang.ERROR_CMD_NOT_SPECIFIED.format(err.key))
        elif isinstance(err, Exception):
            print(Lang.ERROR_GENERAL.format(f"{type(err).__name__} {str(err)}"))
        else:
            print(Lang.ERROR_GENERAL.format(err))
    print()
    if print_help:
        print(Lang.MAIN_HELP_STR.format(sys.argv[0]))
    sys.exit()

def work(pars:Optional[dict] = None, clntID:int = -1, q = None, alt:Optional[Dict[str,str]] = None):
    if pars is not None:
        args = ArgChecker(pars=pars, force_parametric=["gen-veh", "gen-scs", "gen-fcs", "plot"])
    else:
        try:
            args = ArgChecker(force_parametric=["gen-veh", "gen-scs", "gen-fcs", "plot"])
        except Exception as e:
            error_exit(e, True)
    if args.pop_bool("h") or args.pop_bool("help"):
        error_exit(None, True)

    sta_pool = StaPool()
    plg_pool = PluginPool()
    if Path("external_components").exists():
        load_external_components("external_components", plg_pool, sta_pool)

    if args.pop_bool("ls-com"):
        print(Lang.MAIN_LS_TITLE_PLG)
        for key, (_, deps) in plg_pool.GetAllPlugins().items():
            if len(deps) > 0:
                print(f"{key}: {','.join(deps)}")
            else:
                print(key)
        print(Lang.MAIN_LS_TITLE_STA)
        print(",".join(sta_pool.GetAllLogItem()))
        sys.exit()

    from_file = args.pop_str("file", "")
    if from_file != "":  # Read parameters from file
        try:
            with open(from_file, "r") as f:
                command = f.read().strip()
        except Exception as e:
            error_exit(Lang.ERROR_FAIL_TO_OPEN.format(from_file, e))
        try:
            kwargs = get_sim_params(command, plg_pool, sta_pool)
        except Exception as e:
            error_exit(str(e), True)
        visible = False
    else:  # Read parameters from command line
        if platform.system() == "Windows":
            if "show" in args.keys():
                print(Lang.WARN_MAIN_SHOW_MEANINGLESS)
                args.pop_bool("show")
            visible = False
        else:
            visible = args.pop_bool("show")

        no_deamon = args.pop_bool("no-daemon")
        debug_mode = args.pop_bool("debug")
        if debug_mode and not visible:
            print(Lang.WARN_MAIN_DEBUG_MEANINGLESS)
            debug_mode = False

        # Get simulation parameters
        try:
            kwargs = get_sim_params(args, plg_pool, sta_pool)
        except Exception as e:
            error_exit(e, True)

    if visible:
        try:
            from v2simux_gui.progbox import ProgBox
        except:
            print(Lang.WARN_MAIN_GUI_NOT_FOUND)
            visible = False

    kwargs.update({
        "clntID": clntID,
        "mpQ": q,
        "alt_command": alt,
    })
    
    if visible:
        vb = ProgBox(
            ["Driving", "Pending", "Charging", "Parking", "Depleted"],
            "Simulator Dashboard",
        )

        def work():
            try:
                simulate_single(vb=vb, **kwargs)
            except Exception as e:
                if debug_mode:
                    raise e
                logging.exception(e)
            vb.close()

        th = threading.Thread(target=work, daemon=not no_deamon).start()
        vb.mainloop()
    else:
        if clntID == -1: print(Lang.MAIN_SIM_START)
        simulate_single(**kwargs)

if __name__ == "__main__":
    work()


from .uxsim import *
from .utils import *
from .analyzer import *
from .scenario_reader_writer import *

__version__ = "1.10.2"
__author__ = "Toru Seo"
__copyright__ = "Copyright (c) 2023 Toru Seo"
__license__ = "MIT License"



import traceback
from v2simux_gui.com_no_vx import *

import os, sys, subprocess, threading


_ = LangLib.Load(__file__)


class ConvertBox(Tk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title(_("TITLE"))
        self.geometry("600x350")
        
        self.input_folder = StringVar()
        self.output_folder = StringVar()
        
        self.setup_ui()
    
    def setup_ui(self):
        input_frame = Frame(self)
        input_frame.pack(pady=10, padx=20, fill="x")

        Label(input_frame, text=_("INPUT")).pack(anchor="w")

        input_select_frame = Frame(input_frame)
        input_select_frame.pack(fill="x", pady=5)
        
        Entry(input_select_frame, textvariable=self.input_folder).pack(side="left", fill="x", expand=True)
        Button(input_select_frame, text=_("BROWSE"), 
                 command=self.select_input_folder, width=10).pack(side="right", padx=(5, 0))
        
        output_frame = Frame(self)
        output_frame.pack(pady=10, padx=20, fill="x")
        
        Label(output_frame, text=_("OUTPUT")).pack(anchor="w")
        
        output_select_frame = Frame(output_frame)
        output_select_frame.pack(fill="x", pady=5)
        
        Entry(output_select_frame, textvariable=self.output_folder).pack(side="left", fill="x", expand=True)
        Button(output_select_frame, text=_("BROWSE"), 
                 command=self.select_output_folder, width=10).pack(side="right", padx=(5, 0))
        
        partition_frame = Frame(self)
        partition_frame.pack(pady=10, padx=20, fill="x")
        partition_frame.columnconfigure(2, weight=1)

        self.partition_count = StringVar(value="1")
        self.auto_partition = BooleanVar(value=False)

        Label(partition_frame, text=_("PARTITION_COUNT")).grid(row=0, column=0, sticky="w")
        def on_auto_partition_changed(*args):
            state = "disabled" if self.auto_partition.get() else "normal"
            self.entry_count_widget.config(state=state)

        self.entry_count_widget = Spinbox(partition_frame, from_=1, to=32, textvariable=self.partition_count, width=5)
        self.entry_count_widget.grid(row=0, column=1, sticky="w")
        self.check_auto_part_widget = Checkbutton(partition_frame, text=_("PARTITION_TIP"), variable=self.auto_partition)
        self.check_auto_part_widget.grid(row=0, column=2, sticky="e")
        self.auto_partition.trace_add("write", on_auto_partition_changed)

        self.options_frame = Frame(self)
        self.options_frame.pack(pady=10, padx=20, fill="x")

        self.non_passenger_links = BooleanVar(value=False)
        self.check_non_passenger_links = Checkbutton(self.options_frame, text=_("NON_PASSENGER_LINKS"), variable=self.non_passenger_links)
        self.check_non_passenger_links.pack(anchor="w")

        self.non_scc_links = BooleanVar(value=False)
        self.check_non_scc_links = Checkbutton(self.options_frame, text=_("NON_SCC_LINKS"), variable=self.non_scc_links)
        self.check_non_scc_links.pack(anchor="w")

        self.execute_button = Button(self, text=_("EXECUTE"), command=self.execute_program)
        self.execute_button.pack(pady=10)
        
    
    def select_input_folder(self):
        folder = filedialog.askdirectory(title=_("SELECT_INPUT_FOLDER"))
        if folder:
            self.input_folder.set(folder)
    
    def select_output_folder(self):
        folder = filedialog.askdirectory(title=_("SELECT_OUTPUT_FOLDER"))
        if folder:
            self.output_folder.set(str(Path(folder) / Path(self.input_folder.get()).name))
    
    def execute_program(self):
        if not self.input_folder.get() or not self.output_folder.get():
            MB.showwarning(_("WARNING"), _("PLEASE_SELECT_FOLDER"))
            return
        
        self.execute_button.config(state="disabled", text=_("EXECUTING"))
        
        thread = threading.Thread(target=self.run_programs)
        thread.daemon = True
        thread.start()
    
    def run_programs(self):
        result1 = self.execute_first_program()
        self.after(0, self.first_program_completed, result1)
    
    def execute_first_program(self):
        self.input_path = self.input_folder.get()
        self.output_path = self.output_folder.get()

        if not os.path.isdir(self.input_path):
            return False, "", _("INPUT_NOT_EXIST")

        from v2simux import ConvertCase

        try:
            converted = ConvertCase(
                self.input_path, self.output_path, 
                part_cnt = int(self.partition_count.get()),
                auto_partition = self.auto_partition.get(),
                non_passenger_links = self.non_passenger_links.get(),
                non_scc_links = self.non_scc_links.get()
            )
            return converted, "", ""
        except Exception as e:
            traceback.print_exc()
            return False, "", str(e)

    def first_program_completed(self, result):
        ok, out, err = result
        if not ok:
            MB.showerror(_("ERROR"), _("FAILED_MSG").format(err))
            self.reset_ui()
            return
    
        answer = MB.askyesno(_("CONTINUE"), _("CONTINUE_MSG"))
        
        if answer:
            self.destroy()
            import os
            os.system(f'python gui_main.py -d "{self.output_path}"')
        else:
            self.reset_ui()
    
    def reset_ui(self):
        self.execute_button.config(state="normal", text=_("EXECUTE"))

__all__ = ["ConvertBox"]


from v2simux_gui.common import *

from feasytools import RangeList
from .scrtv import ScrollableTreeView, ALWAYS_ONLINE


_L = LangLib.Load(__file__)


class RangeListEditor(Toplevel):
    def __init__(self, data:RangeList, var:StringVar, hint_hms:bool=False):
        super().__init__()
        self.title(_L["RANGE_LIST_EDITOR"])
        self.data = data
        self.tree = ScrollableTreeView(self, allowSave=False)
        self.tree['show'] = 'headings'
        self.tree["columns"] = ("lb", "rb")
        self.tree.column("lb", width=120, stretch=NO)
        self.tree.column("rb", width=120, stretch=NO)
        self.tree.heading("lb", text=_L["LEFT_BOUND"])
        self.tree.heading("rb", text=_L["RIGHT_BOUND"])
        self.tree.pack(fill="both", expand=True)
        for l,r in data:
            self.tree.insert("", "end", values=(l, r))
        self.tree.setColEditMode("lb", ConfigItem(
            name="lb", editor=EditMode.ENTRY, desc="Left Bound", default_value=0))
        self.tree.setColEditMode("rb", ConfigItem(
            name="rb", editor=EditMode.ENTRY, desc="Right Bound", default_value=0))
        if hint_hms:
            self.lb_hint = Label(self, text=_L["TIME_FORMAT"])
            self.lb_hint.pack(fill="x", expand=False)
        self.fr = Frame(self)
        self.fr.pack(fill="x", expand=False)
        self.btn_add = Button(self.fr, text=_L["ADD"], command=self.add, width=6)
        self.btn_add.grid(row=0,column=0,pady=3,sticky="w")
        self.btn_del = Button(self.fr, text=_L["DELETE"], command=self.delete, width=6)
        self.btn_del.grid(row=0,column=1,pady=3,sticky="w")
        self.btn_moveup = Button(self.fr, text=_L["UP"], command=self.moveup, width=6)
        self.btn_moveup.grid(row=0,column=2,pady=3,sticky="w")
        self.btn_movedown = Button(self.fr, text=_L["DOWN"], command=self.movedown, width=6)
        self.btn_movedown.grid(row=0,column=3,pady=3,sticky="w")
        self.btn_save = Button(self.fr, text=_L["SAVE_AND_CLOSE"], command=self.save)
        self.btn_save.grid(row=0,column=4,padx=3,pady=3,sticky="e")
        self.var = var
    
    def add(self):
        self.tree.insert("", "end", values=(0, 0))
    
    def delete(self):
        for i in self.tree.tree.selection():
            self.tree.delete(i)
    
    def moveup(self):
        for i in self.tree.tree.selection():
            p = self.tree.tree.index(i)
            self.tree.tree.move(i, "", p-1)
    
    def movedown(self):
        for i in self.tree.tree.selection():
            p = self.tree.tree.index(i)
            self.tree.tree.move(i, "", p+1)
    
    def save(self):
        try:
            d = self.getAllData()
        except:
            MB.showerror(_L["ERROR"], _L["INVALID_TIME_FORMAT"])
            return
        if len(d) == 0:
            self.var.set(ALWAYS_ONLINE)
        else:
            self.var.set(str(d))
        self.destroy()

    def getAllData(self) -> RangeList:
        res:List[Tuple[int,int]] = []
        for i in self.tree.get_children():
            x = self.tree.tree.item(i, "values")
            res.append((RangeList.parse_time(x[0]), RangeList.parse_time(x[1])))
        return RangeList(res)

__all__ = ["RangeListEditor"]


import v2simux
from v2simux import EV, AllocEnv

def MyAverageAllocFunc(env:AllocEnv, veh_cnt: int, v2g_demand: float, v2g_cap: float):
    """
    Customized V2G Strategy
        env: Environment includes SCS, EVs and current time
        veh_cnt: Number of involved EVs
        v2g_demand: V2G power demanded by grid dispatcher
        v2g_cap: Maximum V2G power output of the SCS
    Returns nothing
    """
    if veh_cnt == 0 or v2g_demand == 0: return
    pd = v2g_demand / veh_cnt

    # Important: Must set the actual V2G power of each involved EV by set_temp_pd(...)
    for ev in env.EVs:
        ev.set_temp_pd(pd)

v2simux.V2GAllocPool.add("MyAverage", MyAverageAllocFunc)

def MyAverageMaxPCAllocator(env: AllocEnv, vcnt:int, max_pc0: float, max_pc_tot: float):
    """
    Average maximum charging power allocator
        env: Allocation environment
        vcnt: Number of vehicles being charged
        max_pc0: Maximum charging power of a single pile, kWh/s
        max_pc_tot: Maximum charging power of the entire CS given by the PDN, kWh/s
    Returns nothing
    """
    if vcnt == 0: return
    pc0 = min(max_pc_tot / vcnt, max_pc0)

    # Important: Must set the actual charging power of each involved EV by set_temp_pc(...)
    for ev in env.EVs:
        ev.set_temp_max_pc(pc0)

v2simux.MaxPCAllocPool.add("MyAverage", MyAverageMaxPCAllocator)

def MyEqualChargeRate(rate: float, ev: EV) -> float:
    """
    Charging power modifier
        rate: Nominal charging power
        ev: EV object
    Returns the actual charging power
    """
    return rate

v2simux.ChargeRatePool.add("MyEqual",MyEqualChargeRate)


import time, json, requests
from dataclasses import dataclass
from typing import Any, Optional, TextIO, Union, overload, List, Dict, Tuple
from pathlib import Path
from ..traffic import CheckFile, DetectFiles, ReadXML


AMAP_KEY_FILE = Path.home() / ".v2simux" / "amap_key.txt"


class CSQueryError(Exception):
    def __init__(self, message:str, obj = None):
        if obj is not None:
            message += f"\nObject: {str(obj)}"
        super().__init__(message)
        self.obj = obj


@dataclass
class CS:
    id: str
    name: str
    lat: float
    lng: float
    
    @staticmethod
    def from_dict(d:Dict[str,Any]) -> 'CS':
        id = d.get('id', '')
        if id == '':
            raise CSQueryError("CS_from_dict: Invalid item id.", d)
        name = d.get('name', '')
        if name == '':
            raise CSQueryError("CS_from_dict: Invalid item name.", d)
        loc_str = d.get('location', '')
        if not isinstance(loc_str, str) or loc_str == '':
            raise CSQueryError("CS_from_dict: Invalid item location.", d)
        loc = loc_str.split(',')
        if len(loc) != 2:
            raise CSQueryError("CS_from_dict: Invalid item location.", d)
        try:
            lat = float(loc[1])
            lng = float(loc[0])
        except (ValueError, IndexError):
            raise CSQueryError("CS_from_dict: Invalid item location.", d)
        return CS(id, name, lat, lng)
    
    @staticmethod
    def from_dict_list(dl:List[Dict[str,Any]], lst:'List[CS]', fh:'Optional[TextIO]' = None):
        for d in dl:
            try:
                cs = CS.from_dict(d)
                lst.append(cs)
            except CSQueryError as e:
                print(e)
            if fh is not None:
                fh.write(f"{cs.id},{cs.name},{cs.lat},{cs.lng}\n")


class Rect:
    @overload
    def __init__(self, lu_lng:float, lu_lat:float, br_lng:float, br_lat:float): ...
    
    @overload
    def __init__(self, lu_lng:str): ...
    
    def __init__(self, lu_lng:Union[str,float], lu_lat:Optional[float]=None, br_lng:Optional[float]=None, br_lat:Optional[float]=None):
        if isinstance(lu_lng, str):
            s = lu_lng.split('|')
            lu = s[0].split(',')
            br = s[1].split(',')
            self.lu_lng = float(lu[0])
            self.lu_lat = float(lu[1])
            self.br_lng = float(br[0])
            self.br_lat = float(br[1])
        elif isinstance(lu_lng, float) and isinstance(lu_lat, float) and isinstance(br_lng, float) and isinstance(br_lat, float):
            self.lu_lng = lu_lng
            self.lu_lat = lu_lat
            self.br_lng = br_lng
            self.br_lat = br_lat
        else:
            raise CSQueryError("Invalid Rect arguments.")
    
    def __str__(self):
        return f"{self.lu_lng:.6f},{self.lu_lat:.6f}|{self.br_lng:.6f},{self.br_lat:.6f}"
    
    def split4(self)->'Tuple[Rect,Rect,Rect,Rect]':
        mid_lng = (self.lu_lng + self.br_lng) / 2
        mid_lat = (self.lu_lat + self.br_lat) / 2
        return (
            Rect(self.lu_lng, self.lu_lat, mid_lng, mid_lat),
            Rect(mid_lng, self.lu_lat, self.br_lng, mid_lat),
            Rect(self.lu_lng, mid_lat, mid_lng, self.br_lat),
            Rect(mid_lng, mid_lat, self.br_lng, self.br_lat)
        )


class AMapPOIReader:
    def __init__(self, key:str, limit:int=100, allyes:bool=False):
        self.key = key
        self.offset = 25
        self.limit = limit
        self.all_yes = allyes
        self.buf = open("buf.csv","w", encoding='utf-8')

    def get_raw(self, rect:Rect, keyword:str) -> Tuple[List[Dict[str,Any]], List[CS]]:
        first_page = self.__get0(rect, keyword, 1)

        # Check info code
        infocode = first_page.get('infocode', None)
        if infocode is None:
            raise CSQueryError("AMapPOIReader_get_raw: No 'infocode' field in response.", first_page)
        if infocode == '10001':
            raise CSQueryError("AMapPOIReader_get_raw: Invalid AMap key.")
        if infocode == '10009':
            raise CSQueryError("AMapPOIReader_get_raw: AMap key does not match the platform.")
        if infocode == '10044':
            raise CSQueryError("AMapPOIReader_get_raw: AMap key usage has reached the limit.")

        # Check page count
        page_count_str = first_page.get('count', None)
        if page_count_str is None:
            raise CSQueryError("AMapPOIReader_get_raw: No 'count' field in response.", first_page)
        try:
            page_count = int(page_count_str)
        except ValueError:
            raise CSQueryError(f"AMapPOIReader_get_raw: Invalid count value: {page_count_str}")

        if page_count > 200:
            print(f"Too many results({page_count}). Splitting the region...")
            if not self.all_yes:
                cont = input("Continue?(Y/N) > ")
                if cont.lower() != 'y':
                    return [], []
            rect_list = rect.split4()
            raw_result = []; parsed_result = []
            for rect in rect_list:
                raw, parsed = self.get_raw(rect, keyword)
                raw_result.extend(raw)
                parsed_result.extend(parsed)
            return raw_result, parsed_result
        
        iterate_num = round(page_count / self.offset) + 1
        print(f"Total items: {page_count}. Queries needed: {iterate_num}")

        if not self.all_yes:
            cont = input("Continue?(Y/N) > ")
            if cont.lower() != 'y':
                return [], []
        
        pois:Optional[List[Dict[str, Any]]] = first_page.get('pois', None)
        if pois is None:
            raise CSQueryError("AMapPOIReader_get_raw: No 'pois' field in response.", first_page)
        raw_result = pois
        parsed_result = []
        CS.from_dict_list(pois, parsed_result, self.buf)

        for i in range(2, iterate_num + 1):
            print(f"\rProgress: {i-1}/{iterate_num}",end="")
            temp_result = self.__get0(rect, keyword, i)
            pois:Optional[List[Dict[str, Any]]] = temp_result.get('pois', None)
            if pois is None:
                continue
            raw_result.extend(pois)
            CS.from_dict_list(pois, parsed_result, self.buf)
            time.sleep(0.35)
        print("\rFinished.               ")

        return raw_result, parsed_result

    def __get0(self, rect:Rect, keyword:str, pagenum:int) -> Dict[str,Any]:
        # 011100(充电站中类)|011102(充换电站)|011103(专用充电站)|073000(电动自行车充电站中类)|073001(电动自行车换电)|073002(电动自行车专用充电站)
        url = f'https://restapi.amap.com/v3/place/polygon?polygon={str(rect)}&keywords={keyword}&offset={self.offset}&page={pagenum}&key={self.key}&types=011100&extensions=all'
        response = requests.get(url)
        response.encoding = 'utf-8'
        result = json.loads(response.text)
        return result


def csQuery(root:str, new_loc:str, ak:str, allyes:bool):
    detects = DetectFiles(root)
    tlbr = None
    if "net" in detects:
        tr = ReadXML(detects["net"]).getroot()
        if tr is None:
            raise RuntimeError(f"Failed to load net file {detects['net']}")
        loc_elem = tr.find("location")
        if loc_elem is not None:
            a,b,c,d = loc_elem.attrib["origBoundary"].split(",")
            tlbr = Rect(float(a),float(b),float(c),float(d))
    if new_loc != "":
        tlbr = Rect(new_loc)
    if tlbr is None:
        raise Exception("No location specified. It can be specified by -p or by the net file.")
    else:
        print(f"Location: {tlbr}")
    reader = AMapPOIReader(ak,allyes = allyes)
    results, cslist = reader.get_raw(tlbr, "充电站")
    
    if len(cslist) == 0: return
    
    print("Saving csv...")
    csv_path = str(Path(root) / "cs.csv")
    CheckFile(csv_path)
    with open(csv_path, 'w', encoding="utf-8") as f:
        f.write("id,name,lat,lng\n")
        for itm in cslist:
            f.write(f"{itm.id},{itm.name},{itm.lat},{itm.lng}\n")

    print("Saving json...")
    json_path = str(Path(root) / "cs.json")
    CheckFile(json_path)
    with open(json_path, 'w', encoding="utf-8") as f:
        json.dump(results, f, ensure_ascii=False)



from v2simux_gui.common import *

from dataclasses import dataclass
from itertools import chain
from feasytools import SegFunc, ConstFunc, TimeFunc, RangeList
from fpowerkit import Bus, Line, Generator, PVWind, ESS, ESSPolicy, Grid as fGrid
from v2simux import RoadNet
from v2simux.traffic.net import Edge, Node
from .prope import PropertyPanel


GenLike = Union[Generator, PVWind, ESS]
PointList = List[Tuple[float, float]]
OESet = Optional[Set[str]]
OAfter = Optional[Callable[[], None]]

def _removesuffix(s:str, suffix:str) -> str:
    if s.endswith(suffix):
        return s[:-len(suffix)]
    return s

@dataclass
class itemdesc:
    type:str
    desc:Any

class BIDC:
    def __init__(self, categories:Iterable[str]):
        self._cls = list(categories)
        self._mp:Dict[int, Tuple[str, Any]] = {}
        self._rv:Dict[Any, int] = {}
    
    @property
    def classes(self):
        return self._cls
    
    def add(self, id:int, cls:str, item:Any):
        if id in self._mp: raise KeyError(f"{id} is already in BIDC")
        self._mp[id] = (cls, item)
        self._rv[item] = id
    
    def pop(self, id:int):
        item = self._mp.pop(id)[1]
        self._rv.pop(item)
    
    def remove(self, item:Any):
        id = self._rv.pop(item)
        self._mp.pop(id)
    
    def get(self, id:int):
        return self._mp[id]
    
    def set_desc(self, id:int, desc:Any):
        cls, _ = self._mp[id]
        self._mp[id] = (cls, desc)
    
    def __getitem__(self, id:int):
        return itemdesc(*self._mp[id])
    
    def __setitem__(self, id:int, val:Union[itemdesc,Tuple[str, Any]]):
        if id in self._mp: self.pop(id)
        if isinstance(val, itemdesc):
            self.add(id, val.type, val.desc)
        elif isinstance(val, tuple):
            assert isinstance(val[0], str)
            self.add(id, val[0], val[1])
        else:
            raise TypeError(f"Invalid val: {val}")
    
    def queryID(self, item:Any):
        return self._rv[item]
    
    def queryIDandCls(self, item:Any):
        id = self._rv[item]
        return id, self._mp[id][0]
    
    def queryCls(self, item:Any):
        id = self._rv[item]
        return self._mp[id][0]
    
class NetworkPanel(Frame):
    def _draw_done_center(self):
        self._center()
        self.__en = True
    
    def __init__(self, master, roadnet:Optional[RoadNet]=None, 
            grid:Optional[fGrid]=None, save_callback:Optional[Callable[[bool],None]]=None, **kwargs):
        super().__init__(master, **kwargs)

        self._Q = EventQueue(self)
        self._Q.register("draw_done", lambda: None)
        self._Q.register("draw_done_center", self._draw_done_center)
        self._Q.register("loaded", lambda: None)
        
        self._item_editing = None
        self._item_editing_id = -1
        self._cv = Canvas(self, bg='white')
        self._cv.pack(side='left',anchor='center',fill=BOTH, expand=1)
        self._cv.bind("<MouseWheel>", self._onMouseWheel)
        self._cv.bind("<Button-1>", self._onLClick)
        self._cv.bind("<Button-3>", self._onRClick)
        self._cv.bind("<B1-Motion>", self._onMotion)
        self._cv.bind("<B3-Motion>", self._onMotion)
        self._cv.bind("<ButtonRelease-1>", self._onRelease)
        self._cv.bind("<ButtonRelease-3>", self._onRelease)

        self._pr = PropertyPanel(self, {}, ConfigItemDict())
        self._pr.tree.AfterFunc = self.__finish_edit
        self._pr.pack(side='right',anchor='e',fill=Y, expand=0)
        self.clear()

        if roadnet is not None:
            self.setRoadNet(roadnet)
        
        if grid is not None:
            self.setGrid(grid)
        
        self.save_callback = save_callback
        self.__saved = True
    
    def savefig(self, save_to:str):
        if save_to.lower().endswith(".eps"):
            self._cv.postscript(file = save_to)
            return
        raise RuntimeError("Only .eps format is supported")
    
    def scale(self, x:float, y:float, s:float, item = 'all'):
        self._cv.scale(item, x, y, s, s)
        self._scale['k'] *= s
        self._scale['x'] = (1 - s) * x + self._scale['x'] * s
        self._scale['y'] = (1 - s) * y + self._scale['y'] * s
    
    def move(self, dx:float, dy:float, item = 'all'):
        self._cv.move(item, dx, dy)
        if item == 'all':
            self._scale['x'] += dx
            self._scale['y'] += dy
    
    def convRealXY2PlotXY(self, realX:float, realY:float):
        return realX * self._scale['k'] + self._scale['x'], (-realY) * self._scale['k'] + self._scale['y']
    
    def convPlotXY2RealXY(self, plotX:float, plotY:float):
        x = (plotX - self._scale['x'])/self._scale['k']
        y = -(plotY - self._scale['y'])/self._scale['k']
        return x, y

    def convLL2PlotXY(self, lon:Optional[float], lat:Optional[float]) -> Tuple[float, float]:
        '''Convert longitude and latitude to canvas coordinates'''
        if lon is None: lon = 0
        if lat is None: lat = 0
        if self._r and self._r.hasGeoProj():
            x, y = self._r.convertLonLat2XY(lon, lat)
        else:
            x, y = lon, lat
        y = -y
        return x * self._scale['k'] + self._scale['x'], y * self._scale['k'] + self._scale['y']
    
    def convPlotXY2LL(self, x:float, y:float) -> Tuple[float, float]:
        '''Convert canvas coordinates to longitude and latitude'''
        x = (x - self._scale['x'])/self._scale['k']
        y = -(y - self._scale['y'])/self._scale['k']
        if self._r is None: return (x, y)
        try:
            return self._r.convertXY2LonLat(x, y)
        except:
            return (x, y)
    
    def clear(self):
        self._cv.delete('all')
        self._scale_cnt = 0
        self._items:BIDC = BIDC(["bus", "bustext", "gen", "gentext", "genconn", "line", "edge", "node"])
        self._Redges:Dict[str, int] = {}
        self._located_edges:Set[str] = set()
        self._drag = {'item': None,'x': 0,'y': 0}
        self._scale = {'k':1.0, 'x':0, 'y':0}
        self._r:Optional[RoadNet] = None
        self._g = None
        self.__en = False
        self._world_colors = None
    
    def setRoadNet(self, roadnet:RoadNet, repaint:bool=True, after:OAfter=None):
        '''
        Set the road network to be displayed
            roadnet: ELGraph, the road network to be displayed
            repaint: bool, whether to repaint the network.
            after: Optional[Callable[[], None]], the function to be called after the network is repainted.
                If after is None, this repaint operation will block the main thread!
                If after is not None, this repaint operation will be done asynchronously.
        '''
        self._r = roadnet
        self._world_colors = self._r.create_color_map()
        if repaint: 
            self._draw_async(after=after)
    
    @property
    def Enabled(self) -> bool:
        return self.__en

    @Enabled.setter
    def Enabled(self, v:bool):
        self.__en = v
    
    @property
    def Grid(self) -> Optional[fGrid]:
        return self._g
    
    def setGrid(self, grid:fGrid, repaint:bool=True):
        '''
        Set the power grid to be displayed
            grid: ELGraph, the road network to be displayed
            repaint: bool, whether to repaint the network.
                This repaint operation will block the main thread!
        '''
        assert isinstance(grid, fGrid)
        self._g = grid
        if repaint: 
            self._draw_async()
    
    def _onLClick(self, event):
        if not self.__en: return
        def _edit_id(typename:str, clicked_item:int) -> int:
            if typename.endswith("conn"): return clicked_item + 1
            elif typename.endswith("text"): return clicked_item + 2
            else: return clicked_item
        x, y = event.x, event.y
        nr_item = self._cv.find_closest(x, y)
        ovl_item = self._cv.find_overlapping(x-5, y-5, x+5, y+5)
        if nr_item and nr_item[0] in ovl_item:
            clicked_item = nr_item[0]
            self.UnlocateAllEdges()
            itm = self._items[clicked_item]
            assert self._r is not None
            assert self._g is not None
            if itm.type == "edge":
                e = self._r.edges[itm.desc]
                node_ids = self._r.node_ids
                self._pr.setData2(
                    (e.id, ConfigItem("Name", EditMode.ENTRY, "Name of the edge")),
                    (e.from_node.id, ConfigItem("From", EditMode.COMBO, "From node", combo_values=node_ids)),
                    (e.to_node.id, ConfigItem("To", EditMode.COMBO, "To node", combo_values=node_ids)),
                    (e.length, ConfigItem("Length", EditMode.ENTRY, "Length in meter\nMay not be the same as the euclidean distance.")),
                    (e.lanes, ConfigItem("Lanes", EditMode.SPIN, "Number of lanes", spin_range=(1,100))),
                    (e.speed_limit, ConfigItem("Speed limit", EditMode.ENTRY, "Speed limit in m/s"))
                )
                self.LocateEdge(itm.desc, 'purple')
                self._item_editing = e
                self._item_editing_id = clicked_item
            elif itm.type == "node":
                n = self._r.nodes[itm.desc]
                if self._r.hasGeoProj():
                    lon, lat = self._r.convertXY2LonLat(n.x, n.y)
                    self._pr.setData2(
                        (n.id, ConfigItem("Name", EditMode.ENTRY, "Name of the node")),
                        (lon, ConfigItem("Longitude", EditMode.ENTRY, "Longitude of the node")),
                        (lat, ConfigItem("Latitude", EditMode.ENTRY, "Latitude of the node")),
                    )
                else:
                    self._pr.setData2(
                        (n.id, ConfigItem("Name", EditMode.ENTRY, "Name of the node")),
                        (n.x, ConfigItem("X", EditMode.ENTRY, "X coordinate of the node")),
                        (n.y, ConfigItem("Y", EditMode.ENTRY, "Y coordinate of the node")),
                    )
                self._item_editing = n
                self._item_editing_id = clicked_item
            elif itm.type in ("bus", "bustext"):
                if itm.type == 'bustext':
                    b = self._g.Bus(_removesuffix(itm.desc,".text"))
                else:
                    b = self._g.Bus(itm.desc)
                if self._r.hasGeoProj():
                    name0, desc0 = "Longitude", "Longitude of the bus"
                    name1, desc1 = "Latitude", "Latitude of the bus"
                else:
                    name0, desc0 = "X", "X coordinate of the bus"
                    name1, desc1 = "Y", "Y coordinate of the bus"
                self._pr.setData2(
                    (b.ID, ConfigItem("Name", EditMode.ENTRY, "Name of the bus")),
                    (b.Lon, ConfigItem(name0, EditMode.ENTRY, desc0)),
                    (b.Lat, ConfigItem(name1, EditMode.ENTRY, desc1)),
                    (b.V, ConfigItem("V/pu", EditMode.ENTRY, "Voltage magnitude in per unit.\nSet to None if not fixed")),
                    (b.MinV, ConfigItem("Vmin/pu", EditMode.ENTRY, "Minimum voltage magnitude in per unit")),
                    (b.MaxV, ConfigItem("Vmax/pu", EditMode.ENTRY, "Maximum voltage magnitude in per unit")),
                    (b.Pd, ConfigItem("Pd/pu", EditMode.SEGFUNC, "Active power demand in per unit")),
                    (b.Qd, ConfigItem("Qd/pu", EditMode.SEGFUNC, "Reactive power demand in per unit")),
                )
                self._item_editing = b
                self._item_editing_id = clicked_item if itm.type == 'bus' else clicked_item + 1
            elif itm.type == "line":
                l = self._g.Line(itm.desc)
                self._pr.setData({
                    "Name":l.ID,"From Bus":l.fBus,"To Bus":l.tBus,
                    "R/pu":l.R,"X/pu":l.X,"MaxI/kA":l.max_I,"Length/km":l.L
                }, ConfigItemDict((
                    ConfigItem("Name", EditMode.ENTRY, "Name of the line"),
                    ConfigItem("From Bus", EditMode.COMBO, "Sending-end bus of the line", combo_values=self._g.BusNames),
                    ConfigItem("To Bus", EditMode.COMBO, "Receiving-end bus of the line", combo_values=self._g.BusNames),
                    ConfigItem("R/pu", EditMode.ENTRY, "Resistance of the line in per unit"),
                    ConfigItem("X/pu", EditMode.ENTRY, "Reactance of the line in per unit"),
                    ConfigItem("MaxI/kA", EditMode.ENTRY, "Thermal limit of the line in kA"),
                    ConfigItem("Length/km", EditMode.ENTRY, "Length of the line in km"),
                )))
                self._item_editing = l
                self._item_editing_id = clicked_item
            elif itm.type in ("gen", "gentext", "genconn"):
                g = self._g.Gen(_removesuffix(_removesuffix(itm.desc,".text"),".conn"))
                self._pr.setData({
                    "Name":g.ID,"Bus":g.BusID,
                    "Longitude":g.Lon,"Latitude":g.Lat,
                    "P/pu":g.P,"Q/pu":g.Q,
                    "Pmax/pu":g.Pmax,"Pmin/pu":g.Pmin,
                    "Qmax/pu":g.Qmax,"Qmin/pu":g.Qmin,
                    "CostA":g.CostA,"CostB":g.CostB,"CostC":g.CostC
                }, ConfigItemDict((
                    ConfigItem("Name", EditMode.ENTRY, "Name of the generator"),
                    ConfigItem("Bus", EditMode.COMBO, "Bus to which the generator is connected", combo_values=self._g.BusNames),
                    ConfigItem("Longitude", EditMode.ENTRY, "Longitude of the generator.\nIf longtitude not available, set it to an X coordinate"),
                    ConfigItem("Latitude", EditMode.ENTRY, "Latitude of the generator.\nIf latitude not available, set it to a Y coordinate"),
                    ConfigItem("P/pu", EditMode.ENTRY, "Active power output in per unit.\nSet to None if not fixed"),
                    ConfigItem("Q/pu", EditMode.ENTRY, "Reactive power output in per unit.\nSet to None if not fixed"),
                    ConfigItem("Pmax/pu", EditMode.SEGFUNC, "Maximum active power output in per unit"),
                    ConfigItem("Pmin/pu", EditMode.SEGFUNC, "Minimum active power output in per unit"),
                    ConfigItem("Qmax/pu", EditMode.SEGFUNC, "Maximum reactive power output in per unit"),
                    ConfigItem("Qmin/pu", EditMode.SEGFUNC, "Minimum reactive power output in per unit"),
                    ConfigItem("CostA", EditMode.SEGFUNC, "Quadratic cost coefficient A in $/(pu pwr·h)^2"),
                    ConfigItem("CostB", EditMode.SEGFUNC, "Quadratic cost coefficient B in $/(pu pwr·h)"),
                    ConfigItem("CostC", EditMode.SEGFUNC, "Quadratic cost coefficient C in $"),
                )))
                self._item_editing = g
                self._item_editing_id = _edit_id(itm.type, clicked_item)
            elif itm.type in ("pvw", "pvwtext", "pvwconn"):
                p = self._g.PVWind(_removesuffix(_removesuffix(itm.desc,".text"), ".conn"))
                self._pr.setData2(
                    (p.ID, ConfigItem("Name", EditMode.ENTRY, "Name of the PV/Wind generator")),
                    (p.BusID, ConfigItem("Bus", EditMode.COMBO, "Bus to which the PV/Wind generator is connected", combo_values=self._g.BusNames)),
                    (p.Lon, ConfigItem("Longitude", EditMode.ENTRY, "Longitude of the PV/Wind generator.\nIf longtitude not available, set it to an X coordinate")),
                    (p.Lat, ConfigItem("Latitude", EditMode.ENTRY, "Latitude of the PV/Wind generator.\nIf latitude not available, set it to a Y coordinate")),
                    (p.P, ConfigItem("P/pu", EditMode.SEGFUNC, "Active power output of the PV/Wind generator")),
                    (p.PF, ConfigItem("Power Factor", EditMode.ENTRY, "Power Factor should be 0.0~1.0")),
                    (p._tag, ConfigItem("Tag", EditMode.COMBO, "Tag should be 'PV' or 'Wind'", combo_values=['PV', 'Wind'])),
                    (p.CC, ConfigItem("Curtail Cost", EditMode.ENTRY, "Unit = $/(pu pwr·h)")),
                )
                self._item_editing = p
                self._item_editing_id = _edit_id(itm.type, clicked_item)
            elif itm.type in ('ess', 'esstext', 'essconn'):
                e = self._g.ESS(_removesuffix(_removesuffix(itm.desc,".text"),".conn"))
                cp = e._cprice
                if cp is not None: cp /= self._g.Sb_kVA
                dp = e._dprice
                if dp is not None: dp /= self._g.Sb_kVA
                self._pr.setData2(
                    (e.ID, ConfigItem("Name", EditMode.ENTRY, "Name of the ESS")),
                    (e.BusID, ConfigItem("Bus", EditMode.COMBO, "Bus to which the ESS is connected", combo_values=self._g.BusNames)),
                    (e.Lon, ConfigItem("Longitude", EditMode.ENTRY, "Longitude of the ESS.\nIf longtitude not available, set it to an X coordinate")),
                    (e.Lat, ConfigItem("Latitude", EditMode.ENTRY, "Latitude of the ESS.\nIf latitude not available, set it to a Y coordinate")),
                    (e.Cap * self._g.Sb_MVA, ConfigItem("Capacity/MWh", EditMode.ENTRY, "Maximum active power output of the ESS")),
                    (e.SOC, ConfigItem("SOC", EditMode.ENTRY, "State of Charge of the ESS, 0~1")),
                    (e.EC, ConfigItem("Ec", EditMode.ENTRY, "Charging Efficiency of the ESS")),
                    (e.ED, ConfigItem("Ed", EditMode.ENTRY, "Discharging Efficiency of the ESS")),
                    (e.MaxPc * self._g.Sb_kVA, ConfigItem("Max Pc/kW", EditMode.ENTRY, "Maximum Charging Power, kW")),
                    (e.MaxPd * self._g.Sb_kVA, ConfigItem("Max Pd/kW", EditMode.ENTRY, "Maximum Discharging Power, kW")),
                    (e.PF, ConfigItem("Power factor", EditMode.ENTRY, "Power factor")),
                    (e._policy.value, ConfigItem("Policy", EditMode.COMBO, "Charging and discharging policy", combo_values=[ESSPolicy.Manual.value,ESSPolicy.Price.value,ESSPolicy.Time.value])),
                    (str(e._ctime), ConfigItem("CTime", EditMode.RANGELIST, "Charging time if policy is time-based")),
                    (str(e._dtime), ConfigItem("DTime", EditMode.RANGELIST, "Discharging time if policy is time-based")),
                    (cp if cp is not None else "None", ConfigItem("CPrice/$/kWh", EditMode.ENTRY, "Charging if price is strictly below than this given price under price-based policy")),
                    (dp if dp is not None else "None", ConfigItem("DPrice/$/kWh", EditMode.ENTRY, "Discharging if price is strictly greater than this given price under price-based policy")),
                )
                self._item_editing = e
                self._item_editing_id = _edit_id(itm.type, clicked_item)
            else:
                self._pr.setDataEmpty()
            self._pr.tree.show_title(f"Type: {itm.type} (ID = {clicked_item})")
            if itm.type in ('bus', 'gen', 'pvw', 'ess', 'node'):
                self._drag['item'] = clicked_item
                self._drag["x"] = event.x
                self._drag["y"] = event.y

    @staticmethod
    def _float2func(v: str):
        v = eval(v)
        if isinstance(v, (float, int)):
            return ConstFunc(v)
        elif isinstance(v, TimeFunc):
            return v
        else:
            return SegFunc(v) # type: ignore

    @property
    def saved(self) -> bool:
        return self.__saved
    @saved.setter
    def saved(self, v:bool):
        if self.save_callback: self.save_callback(v)
        self.__saved = v
    
    def __move_gen(self, i:int, e:GenLike, nLon:float, nLat:float, move_gen:bool=True):
        assert e.Lon is not None and e.Lat is not None
        x0, y0 = self.convLL2PlotXY(e.Lon, e.Lat)
        x1, y1 = self.convLL2PlotXY(nLon, nLat)
        e.Lon = nLon
        e.Lat = nLat
        dx, dy = x1 - x0, y1 - y0
        if move_gen: self._cv.move(i, dx, dy)
        self._cv.move(i-2, dx, dy)
        assert self._g is not None
        self.__replot_genline(i-1, e, self._g.Bus(e.BusID))
    
    def __replot_genline(self, i:int, e:GenLike, b:Bus):
        assert e.Lon is not None and e.Lat is not None
        assert b.Lon is not None and b.Lat is not None
        x0, y0 = self.convLL2PlotXY(e.Lon, e.Lat)
        x1, y1 = self.convLL2PlotXY(b.Lon, b.Lat)
        self._cv.coords(i, x0, y0, x1, y1)
    
    def __move_line(self, i:int, e:Line):
        assert self._g is not None
        latf1, lonf1 = self._g.Bus(e.fBus).position
        assert latf1 is not None and lonf1 is not None
        pf1 = self.convLL2PlotXY(lonf1,latf1)
        latt1, lont1 = self._g.Bus(e.tBus).position
        assert latt1 is not None and lont1 is not None
        pt1 = self.convLL2PlotXY(lont1,latt1)
        self._cv.coords(i, pf1[0], pf1[1], pt1[0], pt1[1])
    
    def __move_edge(self, i:int, e:Edge):
        assert self._r is not None
        (x0, y0), (x1, y1) = self._r.get_offset_shape(e.id)
        (x0, y0) = self.convRealXY2PlotXY(x0, y0)
        (x1, y1) = self.convRealXY2PlotXY(x1, y1)
        self._cv.coords(i, x0, y0, x1, y1)

    def __move_node(self, i:int, e:Node, newPlotX:float, newPlotY:float, move_node:bool=True):
        oldPlotX, oldPlotY = self.convRealXY2PlotXY(e.x, e.y)
        dx, dy = newPlotX - oldPlotX, newPlotY - oldPlotY
        e.x, e.y = self.convPlotXY2RealXY(newPlotX, newPlotY)
        if move_node:
            self._cv.move(i, dx, dy)
        assert self._r is not None
        for l in chain(e.incoming_edges, e.outgoing_edges):
            lid = self._items.queryID(l.id)
            self.__move_edge(lid, l)

    def __move_bus(self, i:int, e:Bus, newLL:Tuple[float, float] = (-1, -1), newXY:Tuple[float, float] = (0, 0), move_bus:bool=True):
        assert e.Lon is not None and e.Lat is not None and self._r is not None
        x0, y0 = self.convLL2PlotXY(e.Lon, e.Lat)
        if newLL != (-1, -1):
            x1, y1 = self.convLL2PlotXY(*newLL)
            e.Lon, e.Lat = newLL
        else:
            x1, y1 = newXY
            e.Lon, e.Lat = self.convPlotXY2LL(x1, y1)
        
        dx, dy = x1 - x0, y1 - y0
        if move_bus:
            self._cv.move(i, dx, dy)
        self._cv.move(i-1, dx, dy)
        assert self._g is not None
        for g in self._g.GensAtBus(e.ID):
            gid = self._items.queryID(g.ID)
            self.__replot_genline(gid-1, g, e)
        for l in chain(self._g._ladjfb[e.ID], self._g._ladjtb[e.ID]):
            lid = self._items.queryID(l.ID)
            self.__move_line(lid, l)

    @staticmethod
    def __chk(s:str):
        s = s.strip().lower()
        if s == "" or s == "none": return None
        else: return s
    
    def __finish_edit(self):
        ret = self._pr.getAllData()
        e = self._item_editing
        i = self._item_editing_id
        
        if isinstance(e, Edge):
            assert self._r is not None
            if ret['Name'] != e.id and ret['Name'] in self._r.edge_ids:
                MB.showerror("Error", f"New name duplicated: {ret['Name']}")
                return
            e.length = float(ret['Length'])
            e.lanes = int(ret['Lanes'])
            e.speed_limit = float(ret['Speed limit'])
            if e.id != ret['Name']:
                self.UnlocateEdge(e.id)
                self._Redges.pop(e.id)
                self._r.rename_edge(e.id, ret['Name'])
                self._items.set_desc(i, ret['Name'])
                self._Redges[e.id] = i
                self.LocateEdge(e.id, 'purple')
            if ret["From"] != e.from_node.id or ret["To"] != e.to_node.id:
                node_ids = set(self._r.node_ids)
                if ret["From"] not in node_ids:
                    MB.showerror("Error", f"Node {ret['From']} does not exist.")
                    return
                if ret["To"] not in node_ids:
                    MB.showerror("Error", f"Node {ret['To']} does not exist.")
                    return
                self._r.update_edge(e.id, ret["From"], ret["To"])
                self.__move_edge(i, e)
        elif isinstance(e, Node):
            assert self._r is not None
            if ret['Name'] != e.id and ret['Name'] in self._r.node_ids:
                MB.showerror("Error", f"New name duplicated: {ret['Name']}")
                return
            if self._r.hasGeoProj():
                nLon = float(ret['Longitude'])
                nLat = float(ret['Latitude'])
                plotX, plotY = self.convLL2PlotXY(nLon, nLat)
            else:
                x, y = float(ret['X']), float(ret['Y'])
                plotX, plotY = self.convRealXY2PlotXY(x, y)
            self.__move_node(i, e, plotX, plotY)
            if e.id != ret['Name']:
                self._r.update_node(e.id, ret['Name'])
                self._items.set_desc(i, ret['Name'])
        elif isinstance(e, Bus):
            assert self._g is not None and self._r is not None
            if ret['Name'] != e.ID and ret['Name'] in self._g.BusNames:
                MB.showerror("Error", f"New name duplicated: {ret['Name']}")
                return
            e.Pd = self._float2func(ret['Pd/pu'])
            e.Qd = self._float2func(ret['Qd/pu'])
            v = self.__chk(ret['V/pu'])
            if v is not None:
                e.fixV(float(v))
            else:
                e.unfixV()
            e.MinV = float(ret['Vmin/pu'])
            e.MaxV = float(ret['Vmax/pu'])
            if self._r.hasGeoProj():
                nLon = float(ret['Longitude'])
                nLat = float(ret["Latitude"])
                self.__move_bus(i, e, newLL=(nLon, nLat))
            else:
                x = float(ret['X'])
                y = float(ret['Y'])
                self.__move_bus(i, e, newXY=(x, y))
            self._g.ChangeBusID(e.ID, ret['Name'])
            self._items.set_desc(i, ret['Name'])
            # e._id = ret['Name']
            self._cv.itemconfig(i-1, text = e.ID)
        elif isinstance(e, Generator):
            assert self._g is not None
            nLon = float(ret['Longitude'])
            nLat = float(ret["Latitude"])
            e.CostA = self._float2func(ret['CostA'])
            e.CostB = self._float2func(ret['CostB'])
            e.CostC = self._float2func(ret['CostC'])
            self._g.ChangeGenBus(e.ID, ret['Bus'])
            self.__move_gen(i, e, nLon, nLat)
            p = self.__chk(ret['P/pu'])
            q = self.__chk(ret['Q/pu'])
            if p is not None:
                e.fixP(eval(p))
            else:
                e.unfixP()
            if q is not None:
                e.fixQ(eval(q))
            else:
                e.unfixQ()
            e.Pmax = self._float2func(ret['Pmax/pu'])
            e.Qmax = self._float2func(ret['Qmax/pu'])
            e.Pmin = self._float2func(ret['Pmin/pu'])
            e.Qmin = self._float2func(ret['Qmin/pu'])
            self._g.ChangeGenID(e.ID, ret['Name'])
            self._items.set_desc(i, ret['Name'])
            # e._id = ret['Name']
        elif isinstance(e, PVWind):
            assert self._g is not None
            nLon = float(ret['Longitude'])
            nLat = float(ret["Latitude"])
            self._g.ChangePVWindBus(e.ID, ret['Bus'])
            self.__move_gen(i, e, nLon, nLat)
            p = self.__chk(ret['P/pu'])
            e.P = self._float2func(p) if p is not None else 0
            self._g.ChangePVWindID(e.ID, ret['Name'])
            self._items.set_desc(i, ret['Name'])
            # e._id = ret['Name']
        elif isinstance(e, Line):
            assert self._g is not None
            self._g.ChangeLineFromBus(e.ID, ret['From Bus'])
            self._g.ChangeLineToBus(e.ID, ret['To Bus'])
            e.R = float(ret['R/pu'])
            e.X = float(ret['X/pu'])
            e.L = float(ret['Length/km'])
            e.max_I = float(ret['MaxI/kA'])
            self.__move_line(i, e)
            self._g.ChangeLineID(e.ID, ret['Name'])
            self._items.set_desc(i, ret['Name'])
            # e._id = ret['Name']
        elif isinstance(e, ESS):
            assert self._g is not None
            nLon = float(ret['Longitude'])
            nLat = float(ret["Latitude"])
            e.Cap = float(ret['Capacity/MWh']) / self._g.Sb_MVA
            e._elec = float(ret['SOC']) * e.Cap
            e.EC = float(ret['Ec'])
            e.ED = float(ret['Ed'])
            e.MaxPc = float(ret['Max Pc/kW']) / self._g.Sb_kVA
            e.MaxPd = float(ret['Max Pd/kW']) / self._g.Sb_kVA
            e._policy = ESSPolicy(ret['Policy'])
            e._ctime = RangeList(eval(ret["CTime"]))
            e._dtime = RangeList(eval(ret["DTime"]))
            cp = ret["CPrice/$/kWh"]
            if cp.lower() == "none": e._cprice = None
            else: e._cprice = float(cp) * self._g.Sb_kVA
            dp = ret["DPrice/$/kWh"]
            if dp.lower() == "none": e._dprice = None
            else: e._dprice = float(dp) * self._g.Sb_kVA
            e.PF = float(ret['Power factor'])
            self._g.ChangeESSBus(e.ID, ret['Bus'])
            self._g.ChangeESSID(e.ID, ret['Name'])
            self._items.set_desc(i, ret['Name'])
            # e._id = ret['Name']
        self.saved = False

    def _onRClick(self, event):
        if not self.__en: return
        self._drag['item'] = 'all'
        self._drag["x"] = event.x
        self._drag["y"] = event.y
    
    def _onMotion(self, event):
        if not self.__en: return
        if self._drag["item"]:
            x, y = event.x, event.y
            dx = x - self._drag["x"]
            dy = y - self._drag["y"]
            self.move(dx, dy, self._drag["item"])
            self._drag["x"] = x
            self._drag["y"] = y
        if isinstance(self._drag["item"],int):
            self.saved = False
    
    def _onRelease(self, event):
        if not self.__en: return
        i = self._drag["item"]
        if isinstance(i,int):
            self.saved = False
            co = self._cv.coords(i)
            if len(co) == 4: 
                x1,y1,x2,y2 = co
                cx = (x1+x2)/2
                cy = (y1+y2)/2
            elif len(co) == 6: # PVW
                x1, y1, x2, y2, x3, y3 = co
                cx = x1
                cy = (y1+y2)/2
            else:
                raise RuntimeError("Invalid item")
            nLon, nLat = self.convPlotXY2LL(cx, cy)
            if self._items[i].type == 'bus':
                assert self._g is not None
                e = self._g.Bus(self._items[i].desc)
                self.__move_bus(i, e, newLL = (nLon, nLat), move_bus = False)
            elif self._items[i].type == 'gen':
                assert self._g is not None
                e = self._g.Gen(self._items[i].desc)
                self.__move_gen(i, e, nLon, nLat, False)
            elif self._items[i].type == 'pvw':
                assert self._g is not None
                e = self._g.PVWind(self._items[i].desc)
                self.__move_gen(i, e, nLon, nLat, False)
            elif self._items[i].type == 'ess':
                assert self._g is not None
                e = self._g.ESS(self._items[i].desc)
                self.__move_gen(i, e, nLon, nLat, False)
            elif self._items[i].type == 'node':
                assert self._r is not None
                e = self._r.nodes[self._items[i].desc]
                self.__move_node(i, e, cx, cy, False)
            self._onLClick(event)
        self._drag["item"] = None
        
    def _onMouseWheel(self, event):
        if not self.__en: return
        if event.delta > 0 and self._scale_cnt < 50:
            s = 1.1
            self._scale_cnt += 1
        elif event.delta < 0 and self._scale_cnt > -50:
            s = 1 / 1.1
            self._scale_cnt -= 1
        else:
            s = 1
        self.scale(event.x, event.y, s)
    
    def _center(self):
        bbox = self._cv.bbox("all")
        if not bbox: return
        cw = bbox[2] - bbox[0]
        ch = bbox[3] - bbox[1]
        ww = self._cv.winfo_width()
        wh = self._cv.winfo_height()
        dx = (ww - cw) / 2 - bbox[0]
        dy = (wh - ch) / 2 - bbox[1]
        self.move(dx, dy)
        s = min(max(ww-50, 100)/cw, max(wh-50, 100)/ch)
        self.scale(ww//2, wh//2, s)
    
    def LocateEdge(self, edge:str, color:str='red'):
        '''Locate an edge by highlighting it in given color, red by default'''
        if edge in self._Redges:
            pid = self._Redges[edge]
            self._cv.itemconfig(pid, fill=color, width=5)
            self._located_edges.add(edge)
    
    def LocateEdges(self, edges:Iterable[str], color:str='red'):
        '''Locate a set of edges by highlighting them in given color, red by default'''
        for edge in edges:
            self.LocateEdge(edge, color)
    
    def UnlocateAllEdges(self):
        '''Unlocate all edges that are located'''
        for edge in self._located_edges:
            self.UnlocateEdge(edge)
        self._located_edges.clear()
    
    def UnlocateEdge(self, edge:str):
        '''Unlocate an edge by restoring its color'''
        if edge in self._Redges:
            pid = self._Redges[edge]
            c, lw = self.__get_edge_prop(edge)
            self._cv.itemconfig(pid, fill=c, width=lw)
        
    def __get_edge_prop(self, edge:str) -> Tuple[Any, float]:
        assert self._r is not None
        # Get the edge properties from the road network
        e = self._r.edges[edge]
        if self._world_colors and e.world_id in self._world_colors:
            return self._world_colors[e.world_id], 2
        return ("blue", 2)

    def _draw_node(self, x:float, y:float, r, color, name:str):
        pid = self._cv.create_oval(x-r, y-r, x+r, y+r, fill=color, width=1)
        self._items[pid] = itemdesc("node", name)

    def _draw_edge(self, shape:PointList, color:str, lw:float, ename:str):
        shape = [(p[0], -p[1]) for p in shape]
        pid = self._cv.create_line(shape, fill=color, width=lw)
        self._items[pid] = itemdesc("edge", ename)
        self._Redges[ename] = pid
    
    def _draw_async(self, scale:float=1.0, dx:float=0.0, dy:float=0.0, center:bool=True, after:OAfter=None):
        if after:
            self._Q.setcallback("draw_done", after)
        self._Q.submit("draw_done", self._draw, scale, dx, dy, center)
    
    def _draw_line(self,x1,y1,x2,y2,color,lw,name):
        self._items[self._cv.create_line(x1,y1,x2,y2,width=lw,fill=color)] = itemdesc('line', name)
    
    def _draw_gen(self,x,y,r,color,lw,name,xb,yb,tp):
        assert tp in ('gen', 'pvw', 'ess')
        self._items[self._cv.create_text(x+1.8*r,y+1.8*r,text=name)] = itemdesc(tp+'text', name+".text")
        self._items[self._cv.create_line(x, y, xb, yb, width=lw)] = itemdesc(tp+"conn", name+".conn")
        if tp == 'gen':
            self._items[self._cv.create_oval(x-r, y-r, x+r, y+r, fill=color, width=lw)] = itemdesc(tp, name)
        elif tp == 'pvw':
            self._items[self._cv.create_polygon(x, y-r, x-r, y+r, x+r, y+r, fill=color, outline='black', width=lw)] = itemdesc(tp, name)
        else:
            self._items[self._cv.create_rectangle(x-r, y-r, x+r, y+r, fill=color, width=lw)] = itemdesc(tp, name)

    def _draw_bus(self,x,y,r,color,lw,name):
        self._items[self._cv.create_text(x+1.8*r,y+1.8*r,text=name)] = itemdesc('bustext', name+".text")
        self._items[self._cv.create_rectangle(x-0.5*r, y-r, x+0.5*r, y+r, fill=color, width=lw)] = itemdesc("bus", name)
    
    def _draw(self, scale:float=1.0, dx:float=0.0, dy:float=0.0, center:bool=True):
        if self._r is None: return
        self.__en = False
        self._cv.delete('all')

        minx, miny, maxx, maxy = self._r.getBoundary()
        if minx > maxx or miny > maxy:
            r = 5
            cx, cy = 0, 0
        else:
            r = max(maxx-minx, maxy-miny)/200
            cx = minx
            cy = miny
            
        for ename in self._r.edges.keys():
            shape = self._r.get_offset_shape(ename)
            c, lw = self.__get_edge_prop(ename)
            shape = [(p[0] * scale + dx, p[1] * scale + dy) for p in shape]
            self._Q.delegate(self._draw_edge, shape, c, lw, ename)
        
        for nd in self._r.nodes.values():
            x, y = nd.get_coord()
            x = x * scale + dx
            y = -y * scale + dy
            self._Q.delegate(self._draw_node, x, y, r/2, 'gray', nd.id)
            
        if self._g is not None:
            locless = 0
            for b in self._g.Buses:
                if b.Lon is None or b.Lat is None:
                    x,y = cx+(locless//20)*7*r, cy+(locless%20)*7*r
                    locless += 1
                    b.Lon, b.Lat = self.convPlotXY2LL(x,y)
                    print(f"Bus {b.ID} has no location, set to Lon, Lat = ({b.Lon:.6f},{b.Lat:.6f})")
            for line in self._g.Lines:
                x1, y1 = self.convLL2PlotXY(*self._g.Bus(line.fBus).LonLat)
                x2, y2 = self.convLL2PlotXY(*self._g.Bus(line.tBus).LonLat)
                self._Q.delegate(self._draw_line, x1, y1, x2, y2, 'black', 2, line.ID)
            
            for g in chain(self._g.Gens, self._g.PVWinds, self._g.ESSs):
                tp = g.__class__.__name__.lower()[:3]
                xb, yb = self.convLL2PlotXY(*self._g.Bus(g.BusID).LonLat)
                if g.Lon is None or g.Lat is None:
                    x,y = xb, yb+3*r
                    locless += 1
                    g.Lon, g.Lat = self.convPlotXY2LL(x,y)
                    t = g.__class__.__name__
                    print(f"{t} {g.ID} has no location, set to Lon, Lat = ({b.Lon:.6f},{b.Lat:.6f})")
                x, y = self.convLL2PlotXY(g.Lon, g.Lat)
                self._Q.delegate(self._draw_gen, x, y, r, 'white', 2, g.ID, xb, yb, tp)

            for b in self._g.Buses:
                x, y = self.convLL2PlotXY(b.Lon, b.Lat)
                self._Q.delegate(self._draw_bus, x, y, r, 'white', 2, b.ID)
        
        if center: self._Q.trigger("draw_done_center")
    
    def save(self, grid_path:str, net_path:str):
        '''Save the current network to files'''
        if self._g:
            self._g.saveFileXML(grid_path)
        if self._r:
            self._r.save(net_path)
        self.saved = True

__all__ = ["NetworkPanel", "itemdesc", "BIDC", "GenLike", "PointList", "OAfter", "OESet"]


from typing import Any, Callable, List, Literal, Optional, Dict, Tuple
from ..locale import Lang
from .ev import EV


_ArriveListener = Callable[[int, EV, Literal[0, 1, 2], float], None]
_ArriveFCSListener = Callable[[int, EV, str, float], None]
_ArriveCSListener = _ArriveFCSListener  # Alias for compatibility
_DepartListener = Callable[[int, EV, int, Optional[str], Any], None]
_DepartDelayListener = Callable[[int, EV, float, int], None]
_DepartFCSListener = Callable[[int, EV, str], None]
_DepartCSListener = _DepartFCSListener  # Alias for compatibility
_JoinSCSListener = Callable[[int, EV, str], None]
_LeaveSCSListener = Callable[[int, EV, str], None]
_DepartFailedListener = Callable[[int, EV, float, str, int], None]
_FaultDepleteListener = Callable[[int, EV, str, int], None]
_FaultNoChargeListener = Callable[[int, EV, str], None]
_FaultRedirectListener = Callable[[int, EV, str, str], None]
_WarnSmallCapListener = Callable[[int, EV, float], None]


class TripsLogger:
    ARRIVAL_NO_CHARGE = 0
    ARRIVAL_CHARGE_SUCCESSFULLY = 1
    ARRIVAL_CHARGE_FAILED = 2
    def __init__(self, file_name:str, append:bool=False):
        self.__ostream = open(file_name, 'a' if append else 'w', encoding='utf-8')
        self.__arrive_listeners:List[_ArriveListener] = []
        self.__arrive_cs_listeners:List[_ArriveFCSListener] = []
        self.__join_scs_listeners:List[_JoinSCSListener] = []
        self.__leave_scs_listeners:List[_LeaveSCSListener] = []
        self.__depart_listeners:List[_DepartListener] = []
        self.__depart_delay_listeners:List[_DepartDelayListener] = []
        self.__depart_cs_listeners:List[_DepartFCSListener] = []
        self.__depart_failed_listeners:List[_DepartFailedListener] = []
        self.__fault_deplete_listeners:List[_FaultDepleteListener] = []
        self.__fault_nocharge_listeners:List[_FaultNoChargeListener] = []
        self.__fault_redirect_listeners:List[_FaultRedirectListener] = []
        self.__warn_smallcap_listeners:List[_WarnSmallCapListener] = []
    
    def add_arrive_listener(self, func: _ArriveListener):
        self.__arrive_listeners.append(func)
    
    def add_arrive_fcs_listener(self, func: _ArriveFCSListener):
        self.__arrive_cs_listeners.append(func)
    
    add_arrive_cs_listener = add_arrive_fcs_listener # Alias for compatibility
    
    def add_depart_listener(self, func: _DepartListener):
        self.__depart_listeners.append(func)
    
    def add_depart_delay_listener(self, func: _DepartDelayListener):
        self.__depart_delay_listeners.append(func)
    
    def add_depart_fcs_listener(self, func: _DepartFCSListener):
        self.__depart_cs_listeners.append(func)
    
    add_depart_cs_listener = add_depart_fcs_listener # Alias for compatibility
    
    def add_join_scs_listener(self, func: _JoinSCSListener):
        self.__join_scs_listeners.append(func)
    
    def add_leave_scs_listener(self, func: _LeaveSCSListener):
        self.__leave_scs_listeners.append(func)

    def add_depart_failed_listener(self, func: _DepartFailedListener):
        self.__depart_failed_listeners.append(func)
    
    def add_fault_deplete_listener(self, func: _FaultDepleteListener):
        self.__fault_deplete_listeners.append(func)
    
    def add_fault_nocharge_listener(self, func: _FaultNoChargeListener):
        self.__fault_nocharge_listeners.append(func)
    
    def add_fault_redirect_listener(self, func: _FaultRedirectListener):
        self.__fault_redirect_listeners.append(func)
    
    def add_warn_smallcap_listener(self, func: _WarnSmallCapListener):
        self.__warn_smallcap_listeners.append(func)
    
    def __pr(self, *args):
        print(*args, file=self.__ostream, sep="|")

    def arrive(self, simT: int, veh: EV, status: Literal[0, 1, 2], dist: float = -1): 
        tid = veh.trip_id
        if tid < veh.trips_count - 1:
            nt = veh.trips[tid + 1]
        else:
            nt = None
        self.__pr(simT, 'A', veh.brief(), status, dist, veh.trip.to_node, nt)
        for l in self.__arrive_listeners:
            l(simT, veh, status, dist)

    def arrive_FCS(self, simT: int, veh: EV, cs: str, dist:float = -1):
        self.__pr(simT, 'AC', veh.brief(), cs, dist)
        for l in self.__arrive_cs_listeners:
            l(simT, veh, cs, dist)

    arrive_CS = arrive_FCS  # Alias for compatibility
    
    def join_SCS(self, simT: int, veh:EV, cs: str):
        self.__pr(simT, 'JS', veh.brief(), cs)
        for l in self.__join_scs_listeners:
            l(simT, veh, cs)
    
    def leave_SCS(self, simT: int, veh: EV, cs: str):
        self.__pr(simT, 'LS', veh.brief(), cs)
        for l in self.__leave_scs_listeners:
            l(simT, veh, cs)

    def depart(self, simT: int, veh: EV, delay:int = 0, cs: Optional[str] = None):
        self.__pr(simT, 'D', veh.brief(), veh.trip, delay, cs, '')
        for l in self.__depart_listeners:
            l(simT, veh, delay, cs, None)

    def depart_delay(self, simT: int, veh: EV, batt_req: float, delay:int):
        self.__pr(simT, 'DD', veh.brief(), veh.battery, batt_req, delay)
        for l in self.__depart_delay_listeners:
            l(simT, veh, batt_req, delay)
    
    def depart_FCS(self, simT: int, veh: EV, cs: str):
        self.__pr(simT, 'DC', veh.brief(), cs, veh.trip.to_node)
        for l in self.__depart_cs_listeners:
            l(simT, veh, cs)
    
    depart_CS = depart_FCS  # Alias for compatibility
    
    def depart_failed(self, simT: int, veh: EV, batt_req: float, cs: str, trT:int):
        self.__pr(simT, 'DF', veh.brief(), veh.battery, batt_req, cs, trT)
        for l in self.__depart_failed_listeners:
            l(simT, veh, batt_req, cs, trT)
    
    def fault_deplete(self, simT: int, veh: EV, cs: str, trT:int):
        self.__pr(simT, 'FD', veh.brief(), cs, trT)
        for l in self.__fault_deplete_listeners:
            l(simT, veh, cs, trT)
    
    def fault_nocharge(self, simT: int, veh: EV, cs: str):
        self.__pr(simT, 'FN', veh.brief(), veh.battery, cs)
        for l in self.__fault_nocharge_listeners:
            l(simT, veh, cs)
    
    def fault_redirect(self, simT: int, veh: EV, cs_old: str, cs_new: str):
        self.__pr(simT, 'FR', veh.brief(), veh.battery, cs_old, cs_new)
        for l in self.__fault_redirect_listeners:
            l(simT, veh, cs_old, cs_new)

    def warn_smallcap(self, simT: int, veh: EV, batt_req: float):
        self.__pr(simT, 'WC', veh.brief(), veh.battery, batt_req)
        for l in self.__warn_smallcap_listeners:
            l(simT, veh, batt_req)
    
    def close(self):
        self.__ostream.close()

    def __del__(self):
        if (hasattr(self, "_TripsLogger__ostream") and
            hasattr(self.__ostream, "closed") and
            not self.__ostream.closed):
            self.__ostream.close()

class TripLogItem:
    OP_NAMEs = {
        'A': Lang.CPROC_ARRIVE,
        'AC': Lang.CPROC_ARRIVE_CS,
        'D': Lang.CPROC_DEPART,
        'DD': Lang.CPROC_DEPART_DELAY,
        'DC': Lang.CPROC_DEPART_CS,
        'DF': Lang.CPROC_DEPART_FAILED,
        'FD': Lang.CPROC_FAULT_DEPLETE,
        'FN': Lang.CPROC_FAULT_NOCHARGE,
        'FR': Lang.CPROC_FAULT_REDIRECT,
        'WC': Lang.CPROC_WARN_SMALLCAP,
        'JS': Lang.CPROC_JOIN_SCS,
        'LS': Lang.CPROC_LEAVE_SCS,
    }
    def __init__(self, simT:int, op:str, veh_id:str, veh_soc:str, veh_batt:str, trip_id:int, additional:Dict[str,str]):
        self.simT = simT
        self.__op = op
        self.veh = veh_id
        self.veh_soc = veh_soc
        self.veh_batt = veh_batt
        self.trip_id = trip_id
        self.additional = additional

    def to_tuple(self,conv:bool=False):
        op = self.__op if not conv else TripLogItem.OP_NAMEs[self.__op]
        return (self.simT, op, self.veh, self.veh_soc, self.veh_batt, self.trip_id, self.additional)
    
    @property
    def time(self):
        """Simulation time in seconds."""
        return self.simT
    
    @property
    def vehicle(self):
        """Vehicle ID."""
        return self.veh
    
    @property
    def soc(self):
        """Vehicle state of charge."""
        return self.veh_soc
    
    @property
    def id(self):
        """Trip ID."""
        return self.trip_id
    
    @property
    def op_raw(self):
        """Operation code."""
        return self.__op

    @property
    def op(self):
        """Operation name, translated if possible."""
        return TripLogItem.OP_NAMEs[self.__op]
    
    @property
    def cs_param(self):
        """Charging station selection parameters, if any."""
        return self.additional.get('cs_param', None)
    
    @property
    def cs(self):
        """Charging station ID, if any."""
        return self.additional.get('cs', None)
    
    def __repr__(self):
        return f"{self.simT}|{self.__op}|{self.veh},{self.veh_soc},{self.veh_batt:.1f},{self.trip_id}|{self.additional}"
    
    def __str__(self):
        ret = f"[{self.simT},{TripLogItem.OP_NAMEs[self.__op]}]"
        veh = f"{self.veh}(Soc={self.veh_soc},TripID={self.trip_id})"
        if self.__op == 'A':
            if self.additional['status'] == '0':
                pos2 = Lang.CPROC_INFO_ARRIVE_0
            elif self.additional['status'] == '1':
                pos2 = Lang.CPROC_INFO_ARRIVE_1
            elif self.additional['status'] == '2':
                pos2 = Lang.CPROC_INFO_ARRIVE_2
            ret += Lang.CPROC_INFO_ARRIVE.format(
                veh, 
                self.additional['arrive_edge'],
                pos2,
                self.additional['next_trip']
            )
        elif self.__op == 'AC':
            ret += Lang.CPROC_INFO_ARRIVE_CS.format(
                veh,
                self.additional['cs']
            )
        elif self.__op == 'D':
            ret += Lang.CPROC_INFO_DEPART.format(
                veh,
                self.additional['trip'],
            )
            if int(self.additional['delay']) > 0:
                ret += Lang.CPROC_INFO_DEPART_WITH_DELAY.format(
                    self.additional['delay']
                )
            if self.additional['cs'] != "None":
                ret += Lang.CPROC_INFO_DEPART_WITH_CS.format(
                    self.additional['cs'],
                    self.additional['cs_param']
                )
        elif self.__op == 'DD':
            ret += Lang.CPROC_INFO_DEPART_DELAY.format(
                veh,
                self.additional['veh_batt'],
                self.additional['batt_req'],
                self.additional['delay']
            )
        elif self.__op == 'DC':
            ret += Lang.CPROC_INFO_DEPART_CS.format(
                veh,
                self.additional['cs'],
                self.additional['arrive_edge']
            )
        elif self.__op == 'DF':
            ret += Lang.CPROC_INFO_DEPART_FAILED.format(
                veh,
                self.additional['veh_batt'],
                self.additional['batt_req'],
                self.additional['cs'],
                self.additional['trT']
            )
        elif self.__op == 'FD':
            ret += Lang.CPROC_INFO_FAULT_DEPLETE.format(
                veh,
                self.additional['cs'],
                self.additional['trT'],
            )
        elif self.__op == 'FN':
            ret += Lang.CPROC_INFO_FAULT_NOCHARGE.format(
                veh,
                self.additional['cs']
            )
        elif self.__op == 'FR':
            ret += Lang.CPROC_INFO_FAULT_REDIRECT.format(
                veh,
                self.additional['old_cs'],
                self.additional['new_cs']
            )
        elif self.__op == 'WC':
            ret += Lang.CPROC_INFO_FAULT_DEPLETE.format(
                veh,
                self.additional['veh_batt'],
                self.additional['batt_req']
            )
        elif self.__op == 'JS':
            ret += Lang.CPROC_INFO_JOIN_SCS.format(
                veh,
                self.additional['cs']
            )
        elif self.__op == 'LS':
            ret += Lang.CPROC_INFO_LEAVE_SCS.format(
                veh,
                self.additional['cs']
            )
        else:
            raise ValueError(f"Unknown operation {self.__op}")
        return ret
    
class TripsReader:
    def __init__(self, filename:str):
        with open(filename, 'r', encoding='utf-8') as fp:
            self.raw_texts = fp.readlines()
        self.meta_data:List[TripLogItem] = []
        self.translated_texts:List[str] = []
        for d in map(lambda x: x.strip().split('|'), self.raw_texts):
            simT = int(d[0])
            op = d[1]
            splits = d[2].split(',')
            if len(splits) == 4:
                veh, soc, batt, tripid = splits
            else:
                veh, soc, tripid = splits
                batt = "None"
            additional:Dict[str,str] = {}
            if op == 'A':
                if len(d) == 6:
                    additional['status'] = d[3]
                    additional['arrive_edge'] = d[4]
                    additional['next_trip'] = d[5]
                elif len(d) == 7:
                    additional['status'] = d[3]
                    additional['dist'] = d[4]
                    additional['arrive_edge'] = d[5]
                    additional['next_trip'] = d[6]
                else:
                    raise ValueError(f"Invalid A entry, expected 6 or 7 fields, got {len(d)}")
            elif op == 'AC':
                if len(d) == 4:
                    additional['cs'] = d[3]
                elif len(d) == 5:
                    additional['cs'] = d[3]
                    additional['dist'] = d[4]
                else:
                    raise ValueError(f"Invalid AC entry, expected 4 or 5 fields, got {len(d)}")
            elif op == 'D':
                assert len(d) == 7, f"Invalid D entry, expected 7 fields, got {len(d)}"
                additional['trip'] = d[3]
                additional['delay'] = d[4]
                additional['cs'] = d[5]
                additional['cs_param'] = d[6]
            elif op == 'DD':
                assert len(d) == 6, f"Invalid DD entry, expected 6 fields, got {len(d)}"
                additional['veh_batt'] = d[3]
                additional['batt_req'] = d[4]
                additional['delay'] = d[5]
            elif op == 'DC':
                assert len(d) == 5, f"Invalid DC entry, expected 5 fields, got {len(d)}"
                additional['cs'] = d[3]
                additional['arrive_edge'] = d[4]
            elif op == 'DF':
                assert len(d) == 7, f"Invalid DF entry, expected 7 fields, got {len(d)}"
                additional['veh_batt'] = d[3]
                additional['batt_req'] = d[4]
                additional['cs'] = d[5]
                additional['trT'] = d[6]
            elif op == 'FD':
                assert len(d) == 5, f"Invalid FD entry, expected 5 fields, got {len(d)}"
                additional['cs'] = d[3]
                additional['trT'] = d[4]
            elif op == 'FN':
                assert len(d) == 4, f"Invalid FN entry, expected 4 fields, got {len(d)}"
                additional['cs'] = d[3]
            elif op == 'FR':
                assert len(d) == 5, f"Invalid FR entry, expected 5 fields, got {len(d)}"
                additional['old_cs'] = d[3]
                additional['new_cs'] = d[4]
            elif op == 'WC':
                assert len(d) == 5, f"Invalid WC entry, expected 5 fields, got {len(d)}"
                additional['veh_batt'] = d[3]
                additional['batt_req'] = d[4]
            elif op == 'JS':
                assert len(d) == 4, f"Invalid JS entry, expected 4 fields, got {len(d)}"
                additional['cs'] = d[3]
            elif op == 'LS':
                assert len(d) == 4, f"Invalid LS entry, expected 4 fields, got {len(d)}"
                additional['cs'] = d[3]
            else:
                raise ValueError(f"Unknown operation {op}")
            met = TripLogItem(simT, op, veh, soc, batt, int(tripid), additional)
            self.meta_data.append(met)
            self.translated_texts.append(str(met))
            
    def __iter__(self):
        return iter(self.translated_texts)
    
    def __len__(self):
        return len(self.translated_texts)
    
    def filter(self, 
        time:Optional[Tuple[Optional[int],Optional[int]]]=None, 
        action:Optional[List[str]]=None, 
        veh:Optional[str]=None, 
        trip_id:Optional[int]=None):
        for r,m,t in zip(self.raw_texts, self.meta_data, self.translated_texts):
            if time is not None:
                if time[0] is not None:
                    if not time[0] <= m.simT:
                        continue
                if time[1] is not None:
                    if not m.simT <= time[1]:
                        continue
            if action is not None:
                if m.op_raw not in action:
                    continue
            if veh is not None:
                if m.veh != veh:
                    continue
            if trip_id is not None:
                if m.trip_id != trip_id:
                    continue
            yield r, m, t

__all__ = ["TripsLogger", "TripLogItem", "TripsReader"]


from v2simux_gui.common import *


_L = LangLib.Load(__file__)


class PlotPad(Frame):
    def __init__(self, master, show_accum:bool=False, useEntry:bool=False, useTotalText:bool=False, **kwargs):
        super().__init__(master, **kwargs)
        if useEntry:
            self.cb = Entry(self)
        else:
            self.cb = Combobox(self)
            self.cb['values'] = []
        self.cb.pack(side='left',padx=3,pady=5)
        self.accum = BooleanVar(self, False)
        if show_accum:
            self.accum.set(True)
            self.cb_accum = Checkbutton(self, text=_L["BTN_TOTAL"] if useTotalText else _L["BTN_ACCUM"], variable=self.accum)
            self.cb_accum.pack(side='left',padx=3,pady=5)
        else:
            self.cb_accum = None
    
    def setValues(self, values:List[str]):
        if isinstance(self.cb, Combobox):
            self.cb['values'] = values
            self.cb.current(0)
    
    def set(self, item:str):
        if isinstance(self.cb, Combobox):
            self.cb.set(item)
        else:
            self.cb.delete(0,END)
            self.cb.insert(0,item)
    
    def get(self):
        return self.cb.get()
    
    def disable(self):
        self.cb['state']=DISABLED
        if self.cb_accum: self.cb_accum['state']=DISABLED
        
    def enable(self):
        self.cb['state']=NORMAL
        if self.cb_accum: self.cb_accum['state']=NORMAL


from abc import abstractmethod, ABC
from collections import deque
from dataclasses import dataclass
from itertools import chain
from typing import Callable, Deque, Iterable, Optional, List, Dict, Set, Tuple, Union
from feasytools import RangeList, makeFunc, OverrideFunc, TimeFunc, ConstFunc, SegFunc
from .ev import EV
from .utils import IntPairList, PriceList


def _get_price_xml(t: TimeFunc, elem:str) -> str:
    if isinstance(t, ConstFunc):
        return f'<{elem}>\n  <item btime="0" price="{t._val}" />\n</{elem}>'
    elif isinstance(t, SegFunc):
        return t.toXML(elem, "item", "btime", "price")
    elif isinstance(t, OverrideFunc):
        return _get_price_xml(t._val, elem)
    else:
        raise ValueError("Unsupported type of TimeFunc")

@dataclass
class AllocEnv:
    cs: 'CS'
    EVs: Iterable[EV]
    CTime: int

V2GAllocator = Callable[[AllocEnv, int, float, float], None]

def _AverageV2GAllocator(env:AllocEnv, veh_cnt: int, v2g_demand: float, v2g_cap: float):
    if veh_cnt == 0 or v2g_demand == 0: return
    pd = v2g_demand / veh_cnt
    for ev in env.EVs:
        ev.set_temp_pd(pd)

class V2GAllocPool:
    """Charging rate correction function pool"""
    _pool:'Dict[str, V2GAllocator]' = {
        "Average":_AverageV2GAllocator, 
    }

    @staticmethod
    def add(name: str, func: V2GAllocator):
        """Add charging rate correction function"""
        V2GAllocPool._pool[name] = func

    @staticmethod
    def get(name: str) -> V2GAllocator:
        """Get charging rate correction function"""
        return V2GAllocPool._pool[name]

MaxPCAllocator = Callable[[AllocEnv, int, float, float], None]

def _AverageMaxPCAllocator(env: AllocEnv, vcnt:int, max_pc0: float, max_pc_tot: float):
    """
    Average maximum charging power allocator
        env: Allocation environment
        vcnt: Number of vehicles being charged
        max_pc0: Maximum charging power of a single pile, kWh/s
        max_pc_tot: Maximum charging power of the entire CS given by the PDN, kWh/s
    """
    if vcnt == 0: return
    pc0 = min(max_pc_tot / vcnt, max_pc0)
    for ev in env.EVs:
        ev.set_temp_max_pc(pc0)

def _PrioritizedMaxPCAllocator(env: AllocEnv, vcnt:int, max_pc0: float, max_pc_tot: float):
    for ev in env.EVs:
        if max_pc_tot > max_pc0:
            ev.set_temp_max_pc(max_pc0)
            max_pc_tot -= max_pc0
        else:
            ev.set_temp_max_pc(max_pc_tot)
            max_pc_tot = 0

def _TimeBasedMaxPCAllocator(env: AllocEnv, vcnt:int, max_pc0: float, max_pc_tot: float):
    loban:List[Tuple[int, EV]] = []
    for ev in env.EVs:
        loban.append((max(0, ev.trip.depart_time - env.CTime), ev))
        # For EVs in FCS, departure time of this trip is smaller than current time. Therefore, the sequence of EVs is held the same as the original.
        # For EVs in SCS, departure time of this trip is larger than current time. Therefore, EVs departed earlier are charged first.
    loban.sort()
    for _, ev in loban:
        if max_pc_tot > max_pc0:
            ev.set_temp_max_pc(max_pc0)
            max_pc_tot -= max_pc0
        else:
            ev.set_temp_max_pc(max_pc_tot)
            max_pc_tot = 0

class MaxPCAllocPool:
    """Charging rate correction function pool"""
    _pool:'Dict[str, MaxPCAllocator]' = {
        "Average":_AverageMaxPCAllocator,
        "Prioritized":_PrioritizedMaxPCAllocator,
        "TimeBased":_TimeBasedMaxPCAllocator,
    }

    @staticmethod
    def add(name: str, func: MaxPCAllocator):
        """Add charging rate correction function"""
        MaxPCAllocPool._pool[name] = func

    @staticmethod
    def get(name: str) -> MaxPCAllocator:
        """Get charging rate correction function"""
        return MaxPCAllocPool._pool[name]
    

class CS(ABC):
    """Charging Station"""

    @abstractmethod
    def __init__(self,
        name: str, slots: int, bus: str, x: float, y: float, offline: Union[IntPairList, RangeList, None],
        max_pc: float, max_pd: float, price_buy: PriceList, price_sell: PriceList,
        pc_alloc: str="Average", pd_alloc: str="Average"
    ):
        """
        Initialize the CS
            name: CS name, also the name of the corresponding node in the network.
            slots: Number of charging piles in the CS.
            bus: The PDN bus to which the CS connects.
            x: The x-coordinate of the CS.
            y: The y-coordinate of the CS.
            offline: Time range when the CS is offline, such as [(start1, end1), (start2, end2), ...]. 
                None means always online.
            max_pc: Each pile's maximum power for charging an EV, kWh/s.
            max_pd: Each pile's maximum power for discharging an EV, kWh/s.
            price_buy: User charging price list, $/kWh.
                The first list is the time range, and the second list is the price,
                such as ([0, 3600, 7200], [1.1, 1.2, 1.1]).
            price_sell: Electricity selling price list, $/kWh.
                The format is the same as price_buy.
                The CS does not support V2G if an empty tuple `()` is passed.
            pc_alloc:
                The method of allocating the maximum charging power to the vehicle.
                The default is "Average", which means that the power is evenly distributed to all vehicles.
            pd_alloc:
                The method of allocating the actual V2G power to the vehicle.
                The default is "Average", which means that the power is evenly distributed to all vehicles.
        """
        self._name: str = name
        self._slots: int = slots
        self._bus: str = bus
        if offline is None:
            self._offline: RangeList = RangeList()
        elif isinstance(offline, RangeList):
            self._offline = offline
        else:
            self._offline: RangeList = RangeList(offline)
        self._manual_offline: Optional[bool] = None
        self._pbuy: OverrideFunc = OverrideFunc(makeFunc(*price_buy))
        self._psell: Optional[OverrideFunc] = (
            None if price_sell == () else OverrideFunc(makeFunc(*price_sell))
        )

        self._pc_lim1: float = max_pc # Maximum charging power of a single pile
        self._pc_limtot: float = float("inf") # Maximum charging power of the entire CS given by the PDN
        self._pc_alloc_str: str = pc_alloc # Charging power allocation method
        self._pc_alloc: MaxPCAllocator = MaxPCAllocPool.get(pc_alloc)
        self._pc_actual: Optional[List[float]] = None # Actual charging power limit allocated to each slot

        self._pd_lim1: float = max_pd # Maximum V2G discharge power of a single pile
        self._pd_alloc_str: str = pd_alloc # V2G power allocation method
        self._pd_alloc: V2GAllocator = V2GAllocPool.get(pd_alloc) # V2G power allocation function
        self._pd_actual: List[float] = [] # Actual V2G power ratio allocated to each slot

        self._cload: float = 0.0
        self._dload: float = 0.0
        self._cur_v2g_cap: float = 0.0
        self._x: float = x
        self._y: float = y
    
    def __repr__(self):
        return f"CS(name='{self._name}',slots={self._slots},pbuy={self._pbuy},psell={self._psell},offline={self._offline})"
    
    def __str__(self):
        return f"CS(name='{self._name}')"
    
    @abstractmethod
    def to_xml(self) -> str: ...

    @property
    def x(self) -> float:
        """X-coordinate of the charging station"""
        return self._x
    
    @property
    def y(self) -> float:
        """Y-coordinate of the charging station"""
        return self._y
    
    @property
    def name(self) -> str:
        """Charging station name"""
        return self._name

    @property
    def slots(self) -> int:
        """Number of charging piles in the charging station"""
        return self._slots

    @property
    def bus(self) -> str:
        """The distribution network bus to which the charging station belongs"""
        return self._bus

    @property
    def pbuy(self) -> OverrideFunc:
        """Vehicle purchase price, $/kWh"""
        return self._pbuy

    @property
    def psell(self) -> OverrideFunc:
        """Electricity selling price, $/kWh"""
        if self._psell is None:
            raise AttributeError("V2G not supported in %s." % self.name)
        return self._psell

    @property
    def supports_V2G(self) -> bool:
        """Check if this charging station supports V2G"""
        return self._psell is not None

    def is_online(self, t: int) -> bool:
        """
        Check if the charging station is available at $t$ seconds.
            t: Time point
        Return:
            True if available, False if not available (fault)
        """
        if self._manual_offline is not None:
            return not self._manual_offline
        return not self._offline.__contains__(t)
    
    def is_offline(self, t: int) -> bool:
        """
        Check if the charging station is unavailable at $t$ seconds.
            t: Time point
        Return:
            True if not available (fault), False if available
        """
        if self._manual_offline is not None:
            return self._manual_offline
        return self._offline.__contains__(t)
    
    def force_shutdown(self):
        """Manually shut down the charging station"""
        self._manual_offline = True
    
    def force_reopen(self):
        """Manually reopen the charging station"""
        self._manual_offline = False
    
    def clear_manual_offline(self):
        """Clear manual shutdown status"""
        self._manual_offline = None

    @abstractmethod
    def add_veh(self, veh_id: str) -> bool:
        """
        Add a vehicle to the charging queue. Wait when the charging pile is insufficient.
            veh_id: Vehicle ID
        Return:
            True if added successfully, False if the vehicle is already charging.
        """
        raise NotImplementedError

    @abstractmethod
    def pop_veh(self, veh_id: str) -> bool:
        """
        Remove the vehicle from the charging queue.
            veh_id: Vehicle ID
        Return:
            True if removed successfully, False if the vehicle does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def __contains__(self, veh_id: str) -> bool:
        pass

    def has_veh(self, veh_id: str) -> bool:
        """
        Check if there is a vehicle with the specified ID.
            veh_id: Vehicle ID
        Return:
            True if exists, False if not exists.
        """
        return self.__contains__(veh_id)

    @abstractmethod
    def __len__(self) -> int:
        """Return the number of vehicles in the charging station"""
        raise NotImplementedError

    @abstractmethod
    def update(
        self, sec: int, cur_time: int, v2g_demand: float
    ) -> List[str]:
        """
        Charge and discharge the EV with the current parameters for sec seconds.
            sec: Seconds
            cur_time: Current time
            v2g_demand: V2G power demanded by the PDN, kWh/s
        Return:
            List of vehicles removed from CS
        """
        raise NotImplementedError

    @abstractmethod
    def is_charging(self, veh_id: str) -> bool:
        """
        Get the charging status of the vehicle. If the vehicle does not exist, a ValueError will be raised.
            veh_id: Vehicle ID
        Return:
            True if charging, False if waiting.
        """
        raise NotImplementedError

    @abstractmethod
    def veh_count(self, only_charging: bool=False) -> int:
        """
        Return the number of vehicles in the charging station.
        When only_charging is True, only the number of vehicles being charged is returned.
        """
        raise NotImplementedError

    @abstractmethod
    def get_V2G_cap(self) -> float:
        """
        Charge and discharge the EV according to the given EV in EVDict, 
        and get the maximum power of V2G under the current situation, unit kWh/s
        """
        raise NotImplementedError
    
    def set_Pc_lim(self, value: float):
        """
        Set the maximum charging power of the charging station
            value: Maximum charging power, kWh/s
        """
        self._pc_limtot = value
    
    @property
    def Pc(self) -> float:
        """Current charging power, kWh/s"""
        return self._cload

    @property
    def Pc_kW(self) -> float:
        """Current charging power, kW, 3600kW = 1kWh/s"""
        return self._cload * 3600

    @property
    def Pc_MW(self) -> float:
        """Current charging power, MW, 3.6MW = 1kWh/s"""
        return self._cload * 3.6

    @property
    def Pd(self) -> float:
        """Current V2G discharge power, kWh/s"""
        return self._dload

    @property
    def Pd_kW(self) -> float:
        """Current V2G discharge power, kW, 3600kW = 1kWh/s"""
        return self._dload * 3600

    @property
    def Pd_MW(self) -> float:
        """Current V2G discharge power, MW, 3.6MW = 1kWh/s"""
        return self._dload * 3.6

    @property
    def Pv2g(self) -> float:
        """Current maximum V2G discharge power, kWh/s"""
        return self._cur_v2g_cap

    @property
    def Pv2g_kW(self) -> float:
        """Current maximum V2G discharge power, kW, 3600kW = 1kWh/s"""
        return self._cur_v2g_cap * 3600

    @property
    def Pv2g_MW(self) -> float:
        """Current maximum V2G discharge power, MW, 3.6MW = 1kWh/s"""
        return self._cur_v2g_cap * 3.6
    
    def vehicles(self):
        """Get an iterator of all vehicles in the charging station"""
        raise NotImplementedError
    
    @abstractmethod
    def averageSOC(self, only_charing:bool = False) -> float:
        """Average SOC of all vehicles in the charging station"""
        raise NotImplementedError


class SCS(CS):
    """Slow Charging Station"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._chi: Set[EV] = set()  # Vehicles being charged
        self._free: Set[EV] = set()  # Vehicles that have been fully charged

    def to_xml(self) -> str:
        """
        Get the XML string of the charging station
        """
        ret = f'<scs name="{self._name}" node="{self._name}" slots="{self._slots}" bus="{self._bus}" ' \
            f'x="{self._x}" y="{self._y}" max_pc="{self._pc_lim1 * 3600:.2f}" max_pd="{self._pd_lim1 * 3600:.2f}" ' \
            f'pc_alloc="{self._pc_alloc_str}" pd_alloc="{self._pd_alloc_str}">\n'
        ret += _get_price_xml(self._pbuy, "pbuy") + "\n"
        if self._psell: ret += _get_price_xml(self._psell, "psell") + "\n"
        if len(self._offline) > 0: ret += self._offline.toXML("offline") + "\n"
        ret += "</scs>"
        return ret
    
    toXML = to_xml
    
    def add_veh(self, veh: EV) -> bool:
        if veh in self._chi or veh in self._free:
            return False
        if len(self._chi) + len(self._free) < self._slots:
            self._chi.add(veh)
            return True
        else:
            return False

    def pop_veh(self, veh: EV) -> bool:
        try:
            self._chi.remove(veh)
        except KeyError:
            try:
                self._free.remove(veh)
            except:
                return False
        return True

    def __contains__(self, veh: EV) -> bool:
        return veh in self._chi or veh in self._free

    def __len__(self) -> int:
        return len(self._chi) + len(self._free)

    def is_charging(self, veh: EV) -> bool:
        return veh in self._chi

    def veh_count(self, only_charging:bool = False) -> int:
        if only_charging:
            return len(self._chi)
        else:
            return len(self._chi) + len(self._free)

    def get_V2G_cap(self, _t:int, /) -> float:
        if self.is_offline(_t): return 0.0
        tot_rate_ava = 0.0
        # Do not check if psell is None due to performance considerations
        v2gp = self._psell(_t) # type: ignore
        for ev in chain(self._chi, self._free):
            if ev.willing_to_v2g(_t, v2gp):
                tot_rate_ava += ev.max_v2g_rate * ev.eta_discharge
        self._cur_v2g_cap = tot_rate_ava
        return tot_rate_ava

    def vehicles(self):
        return chain(self._chi, self._free)
    
    def averageSOC(self, only_charging:bool = True) -> float:
        if only_charging:
            n = len(self._chi)
            if n == 0: return 0.0
            return sum(ev.SOC for ev in self._chi) / n
        else:
            n = len(self._chi) + len(self._free)
            if n == 0: return 0.0
            return sum(ev.SOC for ev in chain(self._chi, self._free)) / n

    def update(
        self, sec: int, cur_time: int, v2g_demand: float
    ) -> List[EV]:
        """
        Charge and discharge the EV with the current parameters for sec seconds.
            sec: Seconds
            cur_time: Current time
            v2g_demand: V2G power demanded by the PDN, kWh/s
        Return:
            List of vehicles removed from CS
        Note:
            Ensure get_V2G_cap() is called before update() in each time step to get the latest V2G capacity.
        """
        # Do nothing when the charging station fails
        if self.is_offline(cur_time):
            self._cload = self._dload = 0
            return []
        
        Wcharge = Wdischarge = 0
        pb = self._pbuy(cur_time)
        
        # Set temporary maximum charging, where set_temp_max_pc is called.
        # If _pc_alloc do not allocate power to a vehicle, the vehicle's maximum charging power is not limited.
        if len(self._chi) > 0:
            self._pc_alloc(
                AllocEnv(self, self._chi, cur_time),
                len(self._chi), self._pc_lim1, self._pc_limtot
            )

        ret: Set[EV] = set()
        for ev in self._chi:
            if ev.willing_to_slow_charge(cur_time, pb):
                # If V2G discharge is in progress, don't charge to full
                Wcharge += ev.charge(sec, pb, ev._esc_rate)
                if v2g_demand > 0:
                    if ev._elec >= ev._bcap * ev._kv2g or ev._elec >= ev._chtar: ret.add(ev)
                else:
                    if ev._elec >= ev._chtar: ret.add(ev)

        self._chi -= ret
        self._free |= ret

        if v2g_demand > 0 and self._cur_v2g_cap > 0: # Ensure that V2G capacity is greater than 0 and there is V2G demand
            assert self._psell is not None, "V2G not supported in %s." % self.name
            ps = self._psell(cur_time)

            # Get vehicles willing to V2G
            v2g_cars = [ev for ev in self._free if ev.willing_to_v2g(cur_time, ps)]
            
            # Allocate V2G power to vehicles, where set_temp_pd is called.
            # If _pd_alloc do not allocate power to a vehicle, the vehicle's discharging power is set to maximum discharging power.
            n = len(v2g_cars)
            if n > 0:
                self._pd_alloc(
                    AllocEnv(self, v2g_cars, cur_time), 
                    len(v2g_cars), v2g_demand, self._cur_v2g_cap
                )
            
            for ev in v2g_cars:
                Wdischarge += ev.discharge(sec, ps)
        
        self._cload = Wcharge / sec
        self._dload = Wdischarge / sec
        return []


class FCS(CS):
    """Fast Charging Station"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._chi: Set[EV] = set()  # Vehicles being charged
        self._buf: Deque[EV] = deque()  # Vehicles in queue

    def to_xml(self) -> str:
        ret = f'<fcs name="{self._name}" node="{self._name}" slots="{self._slots}" bus="{self._bus}" ' \
            f'x="{self._x}" y="{self._y}" max_pc="{self._pc_lim1 * 3600:.2f}" pc_alloc="{self._pc_alloc_str}">\n'
        ret += _get_price_xml(self._pbuy, "pbuy") + "\n"
        if len(self._offline) > 0: ret += self._offline.toXML("offline") + "\n"
        ret += "</fcs>"
        return ret
    
    def add_veh(self, veh: EV) -> bool:
        if veh in self._chi or veh in self._buf:
            return False
        if len(self._chi) < self._slots:
            self._chi.add(veh)
        else:
            self._buf.append(veh)
        return True

    def pop_veh(self, ev: EV) -> bool:
        if ev in self._chi:
            self._chi.remove(ev)
        else:
            try:
                self._buf.remove(ev)
            except:
                return False
        return True

    def __contains__(self, veh: EV) -> bool:
        return veh.ID in self._chi or veh.ID in self._buf

    def is_charging(self, veh: EV) -> bool:
        return veh in self._chi

    def veh_count(self, only_charging:bool = False) -> int:
        if only_charging:
            return len(self._chi)
        else:
            return len(self._chi) + len(self._buf)

    def vehicles(self):
        return chain(self._chi, self._buf)
    
    def averageSOC(self, include_waiting:bool = True) -> float:
        """
        Average SOC of all vehicles in the charging station.
        When include_waiting is True, the average SOC of all vehicles (including those waiting) is returned.
        When include_waiting is False, only the average SOC of vehicles being charged is returned.
        """
        if include_waiting:
            n = len(self._chi) + len(self._buf)
            if n == 0: return 0.0
            return sum(ev.SOC for ev in chain(self._chi, self._buf)) / n
        else:
            n = len(self._chi)
            if n == 0: return 0.0
            return sum(ev.SOC for ev in self._chi) / n

    def wait_count(self) -> int:
        '''Number of vehicles waiting for charging'''
        return len(self._buf)

    def __len__(self) -> int:
        return len(self._chi) + len(self._buf)

    def update(
        self, sec: int, cur_time: int, v2g_demand: float
    ) -> List[EV]:
        """
        Charge the EV with the current parameters for sec seconds.
            sec: Seconds
            cur_time: Current time
            v2g_demand: Useless parameter, ignored. Present for interface consistency only.
        Return:
            List of vehicles removed from CS
        Note:
            Fast charging station does not support V2G.
        """
        Wcharge = 0
        ret:List[EV] = []
        if self.is_offline(cur_time):
            # If the charging station fails, remove all vehicles
            ret = list(chain(self._chi,self._buf))
            self._chi.clear()
            self._buf.clear()
            self._cload = 0
            return ret
        
        # Set temporary maximum charging, where set_temp_max_pc is called.
        # If _pc_alloc do not allocate power to a vehicle, the vehicle's maximum charging power is not limited.
        if len(self._chi) > 0:
            self._pc_alloc(
                AllocEnv(self, self._chi, cur_time), 
                len(self._chi), self._pc_lim1, self._pc_limtot
            )

        for ev in self._chi:
            Wcharge += ev.charge(sec, self.pbuy(cur_time), ev._efc_rate)
            if ev.battery >= ev.charge_target:
                ret.append(ev)
        for ev in ret:
            self.pop_veh(ev)
            if len(self._buf) > 0:
                self._chi.add(self._buf.popleft())
        self._cload = Wcharge / sec
        return ret

    def get_V2G_cap(self, _t:int, /) -> float:
        """
        Fast charging station does not support V2G, always return 0.
        """
        return 0.0



import enum
import os
import sys
import threading
import time
import gzip
import dill as pickle
from collections import deque
from typing import DefaultDict, Deque, Dict, Generator, List, Optional, Set, Tuple, Iterable
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from v2simux.traffic.routing import Link
from .uxsim import World, Vehicle
from .utils import PyVersion, CheckPyVersion
from ..locale.lang import Lang
from .routing import *


class RoutingAlgorithm(enum.Enum):
    AstarFastest = 0
    AstarShortest = 1
    DijkstraFastest = 2
    DijkstraShortest = 3

    def run(self, gl:Graph, coords:CoordsDict, start_time:int, from_node:str, to_node:str):
        if self == RoutingAlgorithm.AstarFastest:
            return astarF(gl, coords, start_time, from_node, to_node)
        elif self == RoutingAlgorithm.AstarShortest:
            return astarS(gl, coords, start_time, from_node, to_node)
        elif self == RoutingAlgorithm.DijkstraFastest:
            return dijF(gl, start_time, from_node, to_node)
        elif self == RoutingAlgorithm.DijkstraShortest:
            return dijS(gl, start_time, from_node, to_node)
        else:
            raise ValueError(Lang.ROUTE_ALGO_NOT_SUPPORTED.format(""))

class WorldSpec(ABC):
    @abstractmethod
    def exec_simulation(self, until_s:int): ...

    @abstractmethod
    def add_vehicle(self, veh_id:str, from_node:str, to_node:str): ...

    @abstractmethod
    def get_arrived_vehicles(self) -> Generator[Tuple[str, Vehicle], None, None]: ...

    @abstractmethod
    def get_time(self) -> int: ...

    @abstractmethod
    def get_gl(self) -> Graph: ...

    @abstractmethod
    def has_vehicle(self, veh_id:str) -> bool: ...

    @abstractmethod
    def get_vehicle(self, veh_id:str) -> Vehicle: ...

    @abstractmethod
    def get_vehicle_count(self) -> int: ...

    @abstractmethod
    def get_coords(self) -> CoordsDict: ...

    @abstractmethod
    def get_average_speed(self) -> float: ...

    @abstractmethod
    def get_link(self, link_id:str) -> Optional[Link]: ...

    @abstractmethod
    def links(self) -> Iterable[Link]: ...

    @abstractmethod
    def shutdown(self): ...

    @abstractmethod
    def _save_obj(self) -> dict: ...
    
    @abstractmethod
    def save(self, filepath:str): ...

    @staticmethod
    @abstractmethod
    def load(filepath:str): ...


def _load_world_unsafe(data: dict) -> WorldSpec:
    assert isinstance(data, dict) and "obj" in data and "version" in data and "pickler" in data, "Invalid world file."
    world_obj = data["obj"]
    assert isinstance(world_obj, WorldSpec), "Invalid world object."
    assert CheckPyVersion(data["version"]), "Incompatible Python version for world: saved {}, current {}".format(data["version"], PyVersion())
    assert data["pickler"] == pickle.__name__, "Incompatible pickler for world: saved {}, current {}".format(data["pickler"], pickle.__name__)
    return world_obj


def _load_world(filepath:str) -> WorldSpec:
    if not os.path.isfile(filepath):
        raise FileNotFoundError(f"File {filepath} does not exist.")
    with gzip.open(filepath, 'rb') as f:
        data = pickle.load(f)
    return _load_world_unsafe(data)


class SingleWorld(WorldSpec):
    def __init__(self, world:World, gl:Graph):
        self.world = world
        self.gl = gl
        self.__ct = 0
        self.__uvi:Dict[str, Vehicle] = {}
        self.__aQ:Deque[str] = deque()
        self.coords:CoordsDict = {}
        self.__cnt = 0
        for node in self.world.NODES:
            self.coords[node.name] = (node.x, node.y)
    
    def exec_simulation(self, until_s:int):
        self.__aQ.clear()
        self.world.exec_simulation(until_s)
        self.__ct = until_s
        self.__cnt += 1
    
    def add_vehicle(self, veh_id:str, from_node:str, to_node:str):
        def __add_to_arrQ(veh:Vehicle):
            self.__aQ.append(veh.name)
        self.__uvi[veh_id] = self.world.addVehicle(orig=from_node, dest=to_node, 
            departure_time=self.__ct, name=veh_id, end_trip_callback=__add_to_arrQ)
    
    def has_vehicle(self, veh_id:str) -> bool:
        return veh_id in self.__uvi
    
    def get_vehicle(self, veh_id:str) -> Vehicle:
        return self.__uvi[veh_id]
    
    def get_vehicle_count(self) -> int:
        return len(self.__uvi)

    def get_running_vehicle_count(self) -> int:
        return len(self.world.VEHICLES_RUNNING)
    
    def get_link(self, link_id:str) -> Optional[Link]:
        return self.world.get_link(link_id)
    
    def links(self) -> List[Link]:
        return self.world.LINKS

    def get_arrived_vehicles(self):
        while len(self.__aQ) > 0:
            veh_id = self.__aQ.popleft()
            yield veh_id, self.__uvi[veh_id]
            self.world.VEHICLES.pop(veh_id)
            del self.__uvi[veh_id]

    def get_time(self) -> int:
        return self.__ct
    
    def get_gl(self) -> Graph:
        return self.gl
    
    def get_coords(self) -> CoordsDict:
        return self.coords
    
    def get_average_speed(self) -> float:
        return self.world.analyzer.average_speed
    
    def shutdown(self):
        return f"Total steps: {self.__cnt}"
    
    def __lstack_save(self, filepath:str):
        sys.setrecursionlimit(10**9)
        with gzip.open(filepath, 'wb') as f:
            pickle.dump({
                "obj": self,
                "version": PyVersion(),
                "pickler": pickle.__name__
            }, f)

    def save(self, filepath:str):
        threading.stack_size(1024 * 1024 * 128)  # 128MB
        t = threading.Thread(target=self.__lstack_save, args=(filepath,))
        t.start()
        t.join()
    
    def _save_obj(self):
        sys.setrecursionlimit(10**9)
        return pickle.dumps({
            "obj": self,
            "version": PyVersion(),
            "pickler": pickle.__name__
        })

    @staticmethod
    def load(filepath:str):
        data = _load_world(filepath)
        assert isinstance(data, SingleWorld)
        return data

class ParaWorlds(WorldSpec):
    def __init__(self, worlds:Dict[int, World], gl:Graph):
        self.worlds = worlds
        self.gl = gl
        self.node_coords:Dict[str, Tuple[float, float]] = {}
        self.wid_of_edges:Dict[str, int] = {}
        self.wid_of_nodes:Dict[str, Set[int]] = DefaultDict(set)
        for wid, W in worlds.items():
            for edge in W.LINKS:
                self.wid_of_edges[edge.name] = wid
            for node in W.NODES:
                self.wid_of_nodes[node.name].add(wid)
                if node.name in self.node_coords:
                    assert self.node_coords[node.name] == (node.x, node.y), \
                        f"Node {node.name} has inconsistent coordinates across worlds."
                else:
                    self.node_coords[node.name] = (node.x, node.y)
        
        # Deque of (arrival vehicle id, current trip segment id)
        self.__aQs:List[Deque[Tuple[str, int]]] = [deque() for _ in range(len(worlds))]

        # Real queue for arrival vehicles, completing the whole trip
        self.__aQ:Deque[str] = deque()

        # time
        self.__ctime = 0

        # Vehicle itineraries: vehicle id -> list of splitting nodes: (node_name, next_world_id)
        self.__veh_itineraies:Dict[str, List[Tuple[str, int]]] = {}

        self.__uvi: Dict[str, Vehicle] = {}

        self.__lt: float = 1.0
        self.__cnt_para = 0
        self.__cnt_ser = 0
        self.__create_pool()
    
    def __create_pool(self):
        self.__pool = ThreadPoolExecutor(os.cpu_count())
    
    def get_coords(self) -> CoordsDict:
        return self.node_coords
    
    def __getitem__(self, wid:int) -> World:
        return self.worlds[wid]

    def get_gl(self) -> Graph:
        return self.gl
    
    def get_arrived_vehicles(self):
        while len(self.__aQ) > 0:
            veh_id = self.__aQ.popleft()
            yield veh_id, self.__uvi[veh_id]
            del self.__uvi[veh_id]
    
    def get_time(self) -> int:
        return self.__ctime

    def exec_simulation(self, until_s:int):
        self.__aQ.clear()
        
        st = time.time()
        if self.__lt < 0.01:
            for W in self.worlds.values():
                W.exec_simulation(until_s)
            self.__cnt_ser += 1
        else:
            futures = []
            for W in self.worlds.values():
                if len(futures) + 1 == len(self.worlds):
                    # The last task in conduct in main thread to reduce the overhead
                    W.exec_simulation(until_s)
                else:
                    futures.append(self.__pool.submit(W.exec_simulation, until_s))
            self.__cnt_para += 1
            for _ in as_completed(futures): pass

        self.__lt = time.time() - st
        
        self.__ctime = until_s

        for i, aQ in enumerate(self.__aQs):
            while len(aQ) > 0:
                veh_id, trip_segment = aQ.popleft()
                self.worlds[i].VEHICLES.pop(veh_id)
                splitting_nodes = self.__veh_itineraies[veh_id]
                if trip_segment + 2 < len(splitting_nodes):
                    trip_segment += 1
                    from_node, next_wid = splitting_nodes[trip_segment]
                    to_node, _ = splitting_nodes[trip_segment + 1]
                    del self.__uvi[veh_id]
                    self.__add_veh(next_wid, veh_id, from_node, to_node, trip_segment)
                else:
                    self.__aQ.append(veh_id)
                    self.__veh_itineraies.pop(veh_id)

    def get_link(self, link_id:str) -> Optional[Link]:
        return self.worlds[self.wid_of_edges[link_id]].get_link(link_id)
    
    def links(self) -> Iterable[Link]:
        return (link for W in self.worlds.values() for link in W.LINKS)

    def __add_veh(self, world_id:int, veh_id:str, from_node:str, to_node:str, trip_segment:int):
        assert world_id in self.wid_of_nodes[from_node], \
            f"Node {from_node} is not in world {world_id}, cannot add vehicle {veh_id}."
        assert world_id in self.wid_of_nodes[to_node], \
            f"Node {to_node} is not in world {world_id}, cannot add vehicle {veh_id}."
        W = self.worlds[world_id]

        def add_to_aQ(veh:Vehicle):
            self.__aQs[world_id].append((veh.name, trip_segment))

        self.__uvi[veh_id] = W.addVehicle(orig=from_node, dest=to_node, 
            departure_time=self.__ctime, name=veh_id, end_trip_callback=add_to_aQ)

    def add_vehicle(self, veh_id:str, from_node:str, to_node:str, algo:RoutingAlgorithm = RoutingAlgorithm.AstarFastest):
        if from_node == to_node:
            for wid in self.wid_of_nodes[from_node]:
                splitting_nodes:List[Tuple[str, int]] = [(from_node, wid), (to_node, -1)]
                break
        else:
            stage = algo.run(self.gl, self.node_coords, self.__ctime, from_node, to_node)
            Ecnt = len(stage.edges)
            assert Ecnt > 0, "Route not found."
            splitting_nodes:List[Tuple[str, int]] = [(stage.nodes[0], self.wid_of_edges[stage.edges[0]])]  # (node_name, prev_world_id, next_world_id)
            for i in range(Ecnt - 1):
                if self.wid_of_edges[stage.edges[i]] != self.wid_of_edges[stage.edges[i + 1]]:
                    splitting_nodes.append((stage.nodes[i + 1], self.wid_of_edges[stage.edges[i + 1]]))
            splitting_nodes.append((stage.nodes[-1], -1))  # Destination node, no next world

        self.__veh_itineraies[veh_id] = splitting_nodes
        self.__add_veh(splitting_nodes[0][1], veh_id, splitting_nodes[0][0], splitting_nodes[1][0], 0)

    def has_vehicle(self, veh_id: str) -> bool:
        return veh_id in self.__uvi
    
    def get_vehicle(self, veh_id: str) -> Vehicle:
        return self.__uvi[veh_id]

    def get_vehicle_count(self) -> int:
        return len(self.__uvi)
    
    def get_running_vehicle_count(self) -> int:
        return sum(len(W.VEHICLES_RUNNING) for W in self.worlds.values())
    
    def get_average_speed(self) -> float:
        return sum(W.analyzer.average_speed for W in self.worlds.values()) / len(self.worlds)
    
    def shutdown(self):
        self.__pool.shutdown(wait=True)
        return f"Total steps: {self.__cnt_ser} serial + {self.__cnt_para} parallel"

    def __lstack_save(self, filepath:str):
        sys.setrecursionlimit(10**9)
        with gzip.open(filepath, 'wb') as f:
            pickle.dump({
                "obj": self,
                "version": PyVersion(),
                "pickler": pickle.__name__
            }, f)

    def save(self, filepath:str):
        self.__pool.shutdown(wait=True)
        del self.__pool
        threading.stack_size(1024 * 1024 * 128)  # 128MB
        t = threading.Thread(target=self.__lstack_save, args=(filepath,))
        t.start()
        t.join()
        self.__create_pool()
    
    def _save_obj(self):
        self.__pool.shutdown(wait=True)
        del self.__pool
        sys.setrecursionlimit(10**9)
        ret = pickle.dumps({
            "obj": self,
            "version": PyVersion(),
            "pickler": pickle.__name__
        })
        self.__create_pool()
        return ret

    @staticmethod
    def load(filepath:str):
        data = _load_world(filepath)
        assert isinstance(data, ParaWorlds), "Invalid world object."
        return data

def load_world(filepath:str) -> WorldSpec:
    data = _load_world(filepath)
    assert isinstance(data, (SingleWorld, ParaWorlds)), "Invalid world object."
    return data

__all__ = ["WorldSpec", "SingleWorld", "ParaWorlds", "RoutingAlgorithm", "load_world", "_load_world_unsafe"]


from .core import (
    TrafficGenerator, DEFAULT_CNAME,
    ProcExisting, ListSelection, PricingMethod,
)
from .csquery import csQuery, AMAP_KEY_FILE
from .tripgen import (
    EVsGenerator, ManualEVsGenerator,
    RoutingCacheMode, TripsGenMode,
)


from collections import defaultdict
from enum import IntEnum
from itertools import repeat
from pathlib import Path
from feasytools import ArgChecker
from fpowerkit import Grid
from sklearn.neighbors import KDTree
from typing import IO, Any, Literal, Tuple, TypeVar, Union, Dict, List
import time
import random

from ..locale import Lang
from ..traffic import DetectFiles, V2SimConfig, RoadNet
from .poly import PolygonMan
from .tripgen import EVsGenerator, RoutingCacheMode, TripsGenMode
from .misc import *

DEFAULT_CNAME = str(Path(__file__).parent.parent / "probtable")

class ProcExisting(IntEnum):
    """How to handle existing files"""

    OVERWRITE = 0  # Overwrite
    SKIP = 1  # Skip
    BACKUP = 2  # Backup
    EXCEPTION = 3  # Raise an exception

    def do(self, path: str):
        if self == ProcExisting.OVERWRITE:
            Path(path).unlink()
        elif self == ProcExisting.SKIP:
            pass
        elif self == ProcExisting.BACKUP:
            i = 0
            while Path(f"{path}.{i}.bak").exists():
                i += 1
            Path(path).rename(f"{path}.{i}.bak")
        else:
            raise FileExistsError(Lang.ERROR_FILE_EXISTS.format(path))

    def check(self, path: str):
        if Path(path).exists():
            self.do(path)


T = TypeVar("T")

class ListSelection(IntEnum):
    """List selection method"""

    ALL = 0  # All
    RANDOM = 1  # Random
    GIVEN = 2  # Given

    def select(self, lst: List[T], n: int = -1, given: List[T] = []) -> List[T]:
        if self == ListSelection.ALL:
            return lst
        elif self == ListSelection.RANDOM:
            if n == -1:
                raise ValueError(Lang.ERROR_NUMBER_NOT_SPECIFIED)
            return random.sample(lst, n)
        else:
            return given


class PricingMethod(IntEnum):
    FIXED = 0  # Fixed price
    RANDOM = 1  # 5-tier random price


class TrafficGenerator:
    def __init__(
        self,
        root: str,
        silent: bool = False,
        existing: ProcExisting = ProcExisting.BACKUP,
    ):
        """
        Generator initialization
            root: Root directory
            silent: Whether to be silent
            existing: How to handle existing files
        """
        self.__root = root
        self.__cfg = DetectFiles(root)
        self.__name = self.__cfg["name"]
        self.__silent = silent
        self.__existing = existing
        self.__start_time = 0
        self.__end_time = 172800
        if self.__cfg.pref:
            pref = V2SimConfig.load(self.__cfg.pref)
            self.__start_time = pref.start_time
            self.__end_time = pref.end_time
        if not self.__cfg.net:
            raise FileNotFoundError(Lang.ERROR_NET_FILE_NOT_SPECIFIED)
        self.__rnet = RoadNet.load(self.__cfg.net)
        self.__ava_fcs: List[str] = [
            e for e in self.__rnet.node_ids if e.upper().startswith("CS")
        ]
        self.__ava_scs: List[str] = [
            e for e in self.__rnet.node_ids if not e.upper().startswith("CS")
        ]
        
        self.__bus_names = ["None"]
        if self.__cfg.grid: 
            self.__bus_names = Grid.fromFile(self.__cfg.grid).BusNames
        else:
            print("Grid is not defined, and thus buses are not included. CS generation may meet errors.")
        
        self.__cs_file = self.__cfg["cscsv"] if "cscsv" in self.__cfg else ""
        self.__grid_file = self.__cfg["grid"] if "grid" in self.__cfg else ""

    def EVTripsFromArgs(self, args: Union[str, ArgChecker]):
        """
        Generate trips from command line arguments
            args: ArgChecker or command line
        """
        if isinstance(args, str):
            args = ArgChecker(args)
        N_cnt = args.pop_int("n")
        day_cnt = args.pop_int("day", 7)
        cname = args.pop_str("c", DEFAULT_CNAME)
        seed = args.pop_int("seed", time.time_ns())
        v2g_prop = args.pop_float("v", 1.0)
        mode_str = args.pop_str("mode", "auto").lower()
        if mode_str == "auto":
            mode = TripsGenMode.AUTO
        elif mode_str == "type":
            mode = TripsGenMode.TYPE
        elif mode_str == "poly":
            mode = TripsGenMode.POLY
        else:
            raise ValueError(Lang.ERROR_INVALID_TRIP_GEN_MODE.format(mode_str))
        cache_route = args.pop_str("cache-route", "none").lower()
        if cache_route == "none":
            rcm = RoutingCacheMode.NONE
        elif cache_route == "runtime":
            rcm = RoutingCacheMode.RUNTIME
        elif cache_route == "static":
            rcm = RoutingCacheMode.STATIC
        else:
            raise ValueError(Lang.ERROR_INVALID_CACHE_ROUTE.format(cache_route))
        if not args.empty():
            raise KeyError(Lang.ERROR_ILLEGAL_CMD.format(','.join(args.to_dict().keys())))
        return self.EVTrips(N_cnt, seed, day_cnt, cname, mode, rcm, v2g_prop=v2g_prop)
    
    def EVTrips(self, n: int, seed: int,
            day_count: int = 7,
            cname: str = DEFAULT_CNAME,
            mode: TripsGenMode = TripsGenMode.AUTO, 
            route_cache: RoutingCacheMode = RoutingCacheMode.NONE,
        **kwargs):
        """
        Generate trips
            n: Number of vehicles
            seed: Randomization seed
            day_count: Number of days
            cname: Trip parameter folder
            mode: Generation mode, "Auto" for automatic, "TAZ" for TAZ-based, "Poly" for polygon-based
            routing_cache: Routing cache mode
            v2g_prop: Proportion of users willing to participate in V2G
            omega: PDFunc | None = None,
            krel: PDFunc | None = None,
            ksc: PDFunc | None = None,
            kfc: PDFunc | None = None,
            kv2g: PDFunc | None = None
        """
        if "veh" in self.__cfg:
            self.__existing.do(self.__cfg["veh"])
        fname = f"{self.__root}/{self.__name}.veh.xml.gz"
        return EVsGenerator(cname, self.__root, seed, mode, route_cache).genEVs(n, fname, day_count, self.__silent, **kwargs)

    def _CS(
        self,
        seed: int,
        *,
        cs_file: str = "",
        poly_file: str = "",
        slots: int = 10,
        mode: Literal["fcs", "scs"] = "fcs",
        bus: ListSelection = ListSelection.ALL,
        busCount: int = -1,
        grid_file: str = "",
        givenBus: List[str] = [],
        cs: ListSelection = ListSelection.ALL,
        csCount: int = -1,
        givenCS: List[str] = [],
        priceBuyMethod: PricingMethod = PricingMethod.FIXED,
        priceBuy: float = 1.0,
        hasSell: bool = False,
        priceSellMethod: PricingMethod = PricingMethod.FIXED,
        priceSell: float = 1.5,
    ):
        def write(fp: IO, price: float):
            loop_end = self.__end_time // 86400
            for d in range(0, loop_end):
                t = [0, 0, 0, 0, 0]
                p = [0.0, 0.0, 0.0, 0.0, 0]
                t[0] = random.choice([0, 1]) if d > 0 else 1
                p[0] = random.uniform(price - 0.5, price - 0.4)
                t[1] = random.choice([6, 7, 8])
                p[1] = random.uniform(price + 0.3, price + 0.6)
                t[2] = random.choice([10, 11])
                p[2] = random.uniform(price - 0.1, price + 0.1)
                t[3] = random.choice([15, 16, 17])
                p[3] = random.uniform(price + 0.2, price + 0.5)
                t[4] = random.choice([19, 20])
                p[4] = random.uniform(price, price + 0.2)
                for tx, px in zip(t, p):
                    fp.write(
                        f'    <item btime="{d*86400+tx*3600}" price="{px:.3}" />\n'
                    )
        warns = []; far_cnt = 0; scc_cnt = 0
        random.seed(seed)
        if mode in self.__cfg:
            self.__existing.do(self.__cfg[mode])
        fname = f"{self.__root}/{self.__name}.{mode}.xml"
        fp = open(fname, "w")
        fp.write("<root>\n")
        cs_pos:Dict[str, Tuple[float, float]] = {}
        if cs_file != "":
            with open(cs_file, "r") as f:
                con = f.readlines()
                _, _, i0, i1 = con[0].strip().split(",")
                if i0 == "lat" and i1 == "lng":
                    swap = False
                elif i0 == "lng" and i1 == "lat":
                    swap = True
                else:
                    raise ValueError("Invalid CSV file.")
                for i in range(1, len(con) - 1):
                    _, _, lat, lng = con[i].strip().split(",")
                    if swap: lat, lng = lng, lat
                    x, y = self.__rnet.convertLonLat2XY(float(lng), float(lat))
                    dist, node = self.__rnet.find_nearest_node_with_distance(x, y)
                    if dist > 200:
                        warns.append(("far_down", lat, lng, x, y, dist))
                        far_cnt += 1
                        continue
                    if not self.__rnet.is_node_in_largest_scc(node.id):
                        warns.append(("scc_down", lat, lng, x, y))
                        scc_cnt += 1
                        continue
                    cs_pos[node.id] = (x, y)
            cs_names = cs.select(sorted(cs_pos.keys()), csCount, givenCS)
            cs_slots = repeat(slots, len(con) - 1)
        elif poly_file != "":
            cs_type:Dict[str,Any] = defaultdict(int)
            PolyMan = PolygonMan(poly_file)
            for poly in PolyMan:
                t = poly.getConvertedType()
                if t is None or t == "Other": continue
                p = poly.center()
                dist, node = self.__rnet.find_nearest_node_with_distance(*p)
                if dist > 200:
                    warns.append(("far_poly", p[0], p[1], dist))
                    far_cnt += 1
                    continue
                if not self.__rnet.is_node_in_largest_scc(node.id):
                    warns.append(("scc_poly", p[0], p[1]))
                    scc_cnt += 1
                    continue
                cs_pos[node.id] = p
                cs_type[node.id] = t
            cs_names = cs.select(sorted(cs_type.keys()), csCount, givenCS)
            def trans(x: str):
                if x == "Home" or x == "Work":
                    return 10 #50
                elif x == "Relax":
                    return 10 #30
                else:
                    raise RuntimeError(f"Invalid type: {x}")
            cs_slots = [trans(cs_type[x]) for x in cs_names]
        else:
            used_cs = self.__ava_fcs if mode == "fcs" else self.__ava_scs
            cs_candidates = []
            for name in used_cs:
                if self.__rnet.is_node_in_largest_scc(name):
                    cs_candidates.append(name)
                else:
                    warns.append(("scc_name", name))
            cs_names = cs.select(cs_candidates, csCount, givenCS)
            cs_slots = repeat(slots, len(cs_names))
            cs_pos = {name: self.__rnet.get_node(name).get_coord() for name in cs_names}
        use_grid = False
        bus_pos:List[Tuple[float, float]] = []
        if grid_file != "":
            gr = Grid.fromFile(grid_file)
            use_grid = True
            for b in gr.Buses:
                lon, lat = b.LonLat
                try:
                    assert lon is not None or lat is not None
                    x, y = self.__rnet.convertLonLat2XY(lon, lat)
                except:
                    use_grid = False
                    break
                bus_pos.append((x, y))
            bus_names = gr.BusNames
        if use_grid:
            bkdt = KDTree(bus_pos, metric="euclidean")
            selector = lambda cname: bus_names[bkdt.query([self.__rnet.get_node(cname).get_coord()], k=1)[1][0][0]]
        else:
            bus_names = bus.select(self.__bus_names, busCount, givenBus)
            selector = lambda cname: random.choice(bus_names)
        for sl, cname in zip(cs_slots, cs_names):            
            fp.write(f'<{mode} name="{mode}_{cname}" node="{cname}" slots="{sl}" bus="{selector(cname)}"')
            if mode == "scs":
                fp.write(f' v2galloc="Average"')
            if cname in cs_pos:
                x, y = cs_pos[cname]
                fp.write(f' x="{x:.1f}" y="{y:.1f}"')
            fp.write(">\n")
            fp.write(f"  <pbuy>\n")
            if priceBuyMethod == PricingMethod.FIXED:
                fp.write(f'    <item btime="0" price="{priceBuy}" />\n')
            else:
                fp.write(f'    <item btime="0" price="1.00" />\n')
                write(fp, priceBuy)
            fp.write(f"  </pbuy>\n")
            if hasSell:
                fp.write(f"  <psell>\n")
                if priceSellMethod == PricingMethod.FIXED:
                    fp.write(f'    <item btime="0" price="{priceSell}" />\n')
                else:
                    fp.write(f'    <item btime="0" price="1.00" />\n')
                    write(fp, priceSell)
                fp.write(f"  </psell>\n")
            fp.write(f"</{mode}>\n")
        fp.write("</root>")
        fp.close()
        return warns, far_cnt, scc_cnt
    
    def FCS(
        self,
        seed: int,
        slots: int,
        *,
        file: str = "",
        bus: ListSelection = ListSelection.ALL,
        busCount: int = -1,
        grid_file: str = "",
        givenBus: List[str] = [],
        cs: ListSelection = ListSelection.ALL,
        csCount: int = -1,
        givenCS: List[str] = [],
        priceBuyMethod: PricingMethod = PricingMethod.FIXED,
        priceBuy: float = 1.5,
    ):
        """
        Generate fast charging station file
            seed: Randomization seed
            slots: Number of charging piles per fast charging station
            bus: Bus selection method
            busCount: Number of buses selected, valid when the bus selection method is random
            givenBus: Specified bus, valid when the bus selection method is specified
            cs: Charging station selection method
            csCount: Number of charging stations selected, valid when the charging station selection method is random
            givenCS: Specified charging station, valid when the charging station selection method is specified
            priceBuyMethod: Pricing method
            priceBuy: Specified price (list)
        """
        return self._CS(
            seed,
            slots = slots,
            mode = "fcs",
            cs_file = file,
            bus=bus,
            busCount=busCount,
            grid_file=grid_file,
            givenBus=givenBus,
            cs=cs,
            csCount=csCount,
            givenCS=givenCS,
            priceBuyMethod=priceBuyMethod,
            priceBuy=priceBuy,
        )
    
    def SCS(
        self,
        seed: int,
        slots: int,
        *,
        file: str = "",
        bus: ListSelection = ListSelection.ALL,
        busCount: int = -1,
        grid_file: str = "",
        givenBus: List[str] = [],
        cs: ListSelection = ListSelection.ALL,
        csCount: int = -1,
        givenCS: List[str] = [],
        priceBuyMethod: PricingMethod = PricingMethod.FIXED,
        priceBuy: float = 1.5,
        priceSellMethod: PricingMethod = PricingMethod.FIXED,
        priceSell: float = 1.5,
    ):
        """
        Generate slow charging station file
            seed: Randomization seed
            slots: Number of charging piles per fast charging station
            bus: Bus selection method
            busCount: Number of buses selected, valid when the bus selection method is random
            givenBus: Specified bus, valid when the bus selection method is specified
            cs: Charging station selection method
            csCount: Number of charging stations selected, valid when the charging station selection method is random
            givenCS: Specified charging station, valid when the charging station selection method is specified
            priceBuyMethod: User purchase price pricing method
            priceBuy: Specified price (list)
            hasSell: Whether to sell electricity
            priceSellMethod: User selling price pricing method
            priceSell: Specified price (list)
        """
        return self._CS(
            seed,
            slots = slots,
            mode = "scs",
            cs_file = file,
            bus=bus,
            busCount=busCount,
            grid_file=grid_file,
            givenBus=givenBus,
            cs=cs,
            csCount=csCount,
            givenCS=givenCS,
            priceBuyMethod=priceBuyMethod,
            priceBuy=priceBuy,
            hasSell=True,
            priceSellMethod=priceSellMethod,
            priceSell=priceSell,
        )
        
    def __CSFromArgs(self, cs_type:str, params: ArgChecker):
        slots = params.pop_int("slots", 10)
        seed = params.pop_int("seed", time.time_ns())
        pbuy = params.pop_float("pbuy", 1.5)
        if self.__cs_file != "":
            cs_file = self.__cs_file
            print("CS file detected: ", cs_file)
        else:
            cs_file = params.pop_str("cs-file", "")
        if self.__grid_file != "":
            grid_file = self.__grid_file
            print("Grid file detected: ", grid_file)
        else:
            grid_file = params.pop_str("grid-file", "")
        randomize_pbuy = params.pop_bool("randomize-pbuy")
        pbuy_method = PricingMethod.RANDOM if randomize_pbuy else PricingMethod.FIXED
        psell = params.pop_float("psell", 1.0)
        randomize_psell = params.pop_bool("randomize-psell")
        psell_method = PricingMethod.RANDOM if randomize_psell else PricingMethod.FIXED
        n_cs = params.pop_int("n-cs", 0)
        cs_names = params.pop_str("cs-names", "").split(",")
        if len(cs_names) == 1 and len(cs_names[0]) == 0:
            cs_names = []
        if n_cs > 0 and len(cs_names) > 0:
            raise Exception(Lang.ERROR_CANNOT_USE_TOGETHER.format("n-cs","cs-names"))
        if n_cs == 0 and len(cs_names) == 0:
            cs_sel = ListSelection.ALL
        elif n_cs > 0:
            cs_sel = ListSelection.RANDOM
        else:
            cs_sel = ListSelection.GIVEN
            
        n_bus = params.pop_int("n-bus", 0)
        new_buses = params.pop_str("bus-names", "").split(",")
        if len(new_buses) == 1 and len(new_buses[0]) == 0:
            new_buses = []
        if n_bus > 0 and len(new_buses) > 0:
            raise Exception(Lang.ERROR_CANNOT_USE_TOGETHER.format("n-bus","bus-names"))
        if n_bus == 0 and len(new_buses) == 0:
            bus_sel = ListSelection.ALL
        elif n_bus > 0:
            bus_sel = ListSelection.RANDOM
        else:
            bus_sel = ListSelection.GIVEN
        
        if cs_type == "fcs":
            self.FCS(seed, slots, file = cs_file, bus=bus_sel, busCount=n_bus, givenBus=new_buses,
                    cs=cs_sel, csCount=n_cs, givenCS=cs_names, priceBuyMethod=pbuy_method, priceBuy=pbuy)
        else:
            self.SCS(seed, slots, file = cs_file, bus=bus_sel, busCount=n_bus, givenBus=new_buses,
                    cs=cs_sel, csCount=n_cs, givenCS=cs_names, priceBuyMethod=pbuy_method, priceBuy=pbuy,
                    priceSellMethod=psell_method, priceSell=psell)

    def CSFromArgs(self, params: Union[str,ArgChecker]):
        if isinstance(params, str):
            params = ArgChecker(params)
        type = params.pop_str("type", "fcs")
        if type not in ["fcs", "scs"]:
            raise Exception(Lang.ERROR_UNKNOWN_CS_TYPE.format(type))
        self.__CSFromArgs(type, params)

    def FCSFromArgs(self, params: Union[str,ArgChecker]):
        if isinstance(params, str):
            params = ArgChecker(params)
        self.__CSFromArgs("fcs", params)
    
    def SCSFromArgs(self, params: Union[str,ArgChecker]):
        if isinstance(params, str):
            params = ArgChecker(params)
        self.__CSFromArgs("scs", params)


from v2simux_gui.common import *

from typing import Protocol, runtime_checkable
from ..mainbox.controls import ScrollableTreeView
from v2simux import ReadOnlyStatistics


_L = LangLib.Load(__file__)


@runtime_checkable
class ISupportSetStatus(Protocol):
    def set_status(self, msg:str): ...


class GridFrame(Frame):
    def __init__(self, parent, to_set, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        self._sta = None
        assert isinstance(to_set, ISupportSetStatus)
        self._master = to_set

        self.panel_time2 = Frame(self)
        self.panel_time2.pack(side='top',fill='x',padx=3,pady=5)
        self.lb_time2 = Label(self.panel_time2, text=_L["TIME_POINT"])
        self.lb_time2.grid(row=0,column=0)
        self.entry_time2 = Entry(self.panel_time2)
        self.entry_time2.insert(0,"86400")
        self.entry_time2.grid(row=0,column=1,sticky='ew')
        self.btn_time2 = Button(self.panel_time2, text=_L["GRID_COLLECT"], takefocus=False, command=self.collectgrid)
        self.btn_time2.grid(row=0,column=2)

        self.grbus = ScrollableTreeView(self)
        self.grbus.pack(side='top',fill='both',padx=3,pady=5)
        self.grbus["show"]="headings"
        self.grbus["columns"]=("bus","v","pd","qd","pg","qg")
        self.grbus.heading("bus",text="Bus")
        self.grbus.heading("v",text="Voltage/kV")
        self.grbus.heading("pd",text="Active load/MW")
        self.grbus.heading("qd",text="Reactive load/Mvar")
        self.grbus.heading("pg",text="Active gen/MW")
        self.grbus.heading("qg",text="Reactive gen/Mvar")
        self.grbus.column("bus",width=50)
        self.grbus.column("v",width=90)
        self.grbus.column("pd",width=100)
        self.grbus.column("qd",width=100)
        self.grbus.column("pg",width=100)
        self.grbus.column("qg",width=100)
        
        self.grline = ScrollableTreeView(self)
        self.grline.pack(side='top',fill='both',padx=3,pady=5)
        self.grline["show"]="headings"
        self.grline["columns"]=("line","p","q","i")
        self.grline.heading("line",text="Line")
        self.grline.heading("p",text="Active pwr/MW")
        self.grline.heading("q",text="Reactive pwr/Mvar")
        self.grline.heading("i",text="Current/kA")
        self.grline.column("line",width=50)
        self.grline.column("p",width=100)
        self.grline.column("q",width=100)
        self.grline.column("i",width=100)

    def setSta(self, sta:ReadOnlyStatistics):
        self._sta = sta
    
    def collectgrid(self):
        self.grbus.clear()
        try:
            t = int(self.entry_time2.get())
        except:
            self._master.set_status("Invalid time point!")
            return
        if self._sta is None:
            self._master.set_status("Statistics not loaded!")
            return
        for b in self._sta.bus_head:
            v = self._sta.bus_attrib_of(b, "V").value_at(t)
            pd = self._sta.bus_attrib_of(b, "Pd").value_at(t)
            qd = self._sta.bus_attrib_of(b, "Qd").value_at(t)
            pg = self._sta.bus_attrib_of(b, "Pg").value_at(t)
            qg = self._sta.bus_attrib_of(b, "Qg").value_at(t)
            self.grbus.insert("",'end',values=(b,v,pd,qd,pg,qg))
        for l in self._sta.line_head:
            p = self._sta.line_attrib_of(l, "P").value_at(t)
            q = self._sta.line_attrib_of(l, "Q").value_at(t)
            i = self._sta.line_attrib_of(l, "I").value_at(t)
            self.grline.insert("",'end',values=(l,p,q,i))
        self._master.set_status(_L["STA_READY"])